#include "px/px.hh"
#include "sdlutil.hh"
#include "mesh.hh"
#include "framerate.hh"

#include <cstdio>
#include <cstdlib>
#include <cstdint>
#include <cstring>
#include <cmath>
#include <cassert>

#include <queue>
#include <vector>
#include <unordered_map>

using namespace px;

// key-bindings based on compile-time dvorak setting

#ifndef DVORAK

#define FORWARD_KEY  SDLK_w
#define LEFT_KEY     SDLK_a
#define BACK_KEY     SDLK_s
#define RIGHT_KEY    SDLK_d

#else

#define FORWARD_KEY  SDLK_COMMA
#define LEFT_KEY     SDLK_a
#define BACK_KEY     SDLK_o
#define RIGHT_KEY    SDLK_e

#endif

/**********************************************************************/

PX_REC_DEF_FIELD(texcoord);
PX_REC_DEF_FIELD(projection);
PX_REC_DEF_FIELD(modelview);
PX_REC_DEF_FIELD(texture);
PX_REC_DEF_FIELD(color);
PX_REC_DEF_FIELD(portalmask);
PX_REC_DEF_FIELD(portaldepth);

// context/shader_obj types
typedef record<L_projection_t, mat4,
        record<L_modelview_t,  mat4,
        record<L_texture_t,    buffer<vec3>*,
        record<L_portaldepth_t, buffer<float>*, empty_rec_t> > > >
    Uni;

typedef record<L_texcoord_t, vec2, empty_rec_t>
    VAttr;

typedef Uint32 SDL_Px;

typedef record<L_color_t,      SDL_Px,
        record<L_portalmask_t, bool, empty_rec_t> >
    Px;

/**********************************************************************/

struct model {
    mesh geom;
    std::unique_ptr<buffer<vec3> > texture;

    bool load(const char* path_prefix, const char** err) {
        char fname[128];
        *err = "";

        snprintf(fname, sizeof (fname), "%s.obj", path_prefix);
        if (!geom.load(fname)) {
            *err = geom.error;
            return false;
        }

        snprintf(fname, sizeof (fname), "%s.png", path_prefix);
        texture = load_rgb(fname, err);
        if (texture == NULL) {
            geom.unload();
            return false;
        }

        return true;
    }

    void draw(context<Uni, VAttr, Px>& ctx) const {
        ctx.uniform()[L_texture] = texture.get();
        ctx.draw_tris(geom.nverts, geom.verts,
                      geom.ntris, geom.tris,
                      (L_texcoord = geom.texcoords));
    }
};

/**********************************************************************/

struct portal_results {
    bool    hit;
    float_4 min_x, min_y, max_x, max_y;

    void reset() {
        float_4 inf(std::numeric_limits<float>::infinity());
        hit = false;
        min_x = inf;
        min_y = inf;
        max_x = -inf;
        max_y = -inf;
    }
};

void draw_portal(model* pmod, context<Uni, VAttr, Px>& ctx,
                 portal_results& pres, shader_obj_id portal_shader) {
    // save
    bool was_wireframe       = ctx.wireframe();
    bool was_cull            = ctx.cull_enabled();
    cull_face_t cull_face    = ctx.cull_face();
    shader_obj_id old_shader = ctx.current_shader();

    // set up
    ctx.set_wireframe(false);
    ctx.enable_cull(true);
    ctx.set_cull_face(CULL_BACK);
    ctx.bind_shader(portal_shader);

    // draw
    pres.reset();
    pmod->draw(ctx);

    // restore
    ctx.set_wireframe(was_wireframe);
    ctx.enable_cull(was_cull);
    ctx.set_cull_face(cull_face);
    ctx.bind_shader(old_shader);
}

/**********************************************************************/

struct room_models {
    std::vector<model*>      parts;
    std::vector<model*>      portal_models;
    std::vector<std::string> portal_rooms;
    std::vector<int>         portal_match;
    std::vector<vec2>        portal_center;
    char error[256];

    ~room_models() {
        for (auto it = parts.begin(); it != parts.end(); it++) {
            if (*it != NULL) {
                delete *it;
                *it = NULL;
            }
        }
        parts.clear();
    }

    bool load(const char* room_path) {
        char fname[128];

        // read room.list
        std::vector<std::string> model_paths;
        snprintf(fname, sizeof (fname), "%s/room.list", room_path);
        FILE* fp = fopen(fname, "r");
        if (fp == NULL) {
            snprintf(error, sizeof (error), "%s: %s", fname, strerror(errno));
            return false;
        }

        char line[128];
        while (fgets(line, sizeof (line), fp) != NULL) {
            int len = strlen(line);
            if (line[len-1] == '\n')
                line[--len] = '\0';

            if (line[0] != '\0') // skip blank lines
                model_paths.push_back(line);
        }

        fclose(fp);

        // load each model
        for (auto it = model_paths.begin(); it != model_paths.end(); it++) {
            if (strstr(it->c_str(), ":") != NULL)
                continue; // portal line

            model* mod = new model;
            const char* errmsg;
            snprintf(fname, sizeof (fname), "%s/%s", room_path, it->c_str());
            if (mod->load(fname, &errmsg)) {
                parts.push_back(mod);
            } else {
                snprintf(error, sizeof (error), "%s", errmsg);
                delete mod;
                for (auto itp = parts.begin(); itp != parts.end(); itp++) {
                    delete *itp;
                    *itp = NULL;
                }
                parts.clear();
                return false;
            }
        }

        // load each portal
        for (auto it = model_paths.begin(); it != model_paths.end(); it++) {
            char buf[512];
            snprintf(buf, sizeof (buf), "%s", it->c_str());

            char* other;
            for (other = buf; *other != '\0' && *other != ':'; other++) { }
            if (*other == '\0')
                continue;

            *other++ = '\0';
            char* other_i;
            for (other_i = other; *other_i != '\0' && *other_i != ':'; other_i++) { }
            if (*other_i == '\0') {
                // it's too late for error reporting
                other_i = (char*)"0";
            } else {
                *other_i++ = '\0';
            }

            char* center_x;
            for (center_x = other_i; *center_x != '\0' && *center_x != ':'; center_x++) { }
            if (*center_x == '\0') {
                center_x = (char*)"0";
            } else {
                *center_x++ = '\0';
            }

            char* center_z;
            for (center_z = center_x; *center_z != '\0' && *center_z != ':'; center_z++) { }
            if (*center_z == '\0') {
                center_z = (char*)"0";
            } else {
                *center_z++ = '\0';
            }

            model* mod = new model;
            const char* errmsg;
            snprintf(fname, sizeof (fname), "%s/%s", room_path, buf);
            if (mod->load(fname, &errmsg)) {
                portal_models.push_back(mod);
                portal_rooms.push_back(std::string(other));
                portal_match.push_back(atoi(other_i));
                portal_center.push_back(vec2(strtof(center_x, NULL), strtof(center_z, NULL)));
            } else {
                snprintf(error, sizeof (error), "%s", errmsg);
                delete mod;
                for (auto itp = parts.begin(); itp != parts.end(); itp++) {
                    delete *itp;
                    *itp = NULL;
                }
                for (auto itp = portal_models.begin();
                     itp != portal_models.end(); itp++) {
                    delete *itp;
                    *itp = NULL;
                }
                parts.clear();
                portal_models.clear();
                portal_rooms.clear();
                portal_match.clear();
                portal_center.clear();
                return false;
            }
        }

        return true;
    }

    void draw(context<Uni, VAttr, Px>& ctx,
              portal_results& pres,
              const std::unordered_map<std::string, room_models*>& rooms,
              unsigned from_portal,
              shader_obj_id portal_shader) const {

        draw_parts(ctx);

        for (unsigned i = 0; i < portal_models.size(); i++) {
            if (i == from_portal)
                continue;

            draw_portal(portal_models[i], ctx, pres, portal_shader);
            if (!pres.hit)
                continue;

            room_models* other_room   = rooms.at(portal_rooms[i]);
            int          other_port_i = portal_match[i];
            model*       other_port   = other_room->portal_models[other_port_i];

            mesh& my_geom = portal_models[i]->geom;
            mesh& op_geom = other_port->geom;

            vec2 my_mid = portal_center[i];
            vec2 op_mid = other_room->portal_center[other_port_i];

            vec3 my_a = my_geom.verts[my_geom.tris[0]];
            vec3 my_b = my_geom.verts[my_geom.tris[1]];
            vec3 my_c = my_geom.verts[my_geom.tris[2]];

            vec3 op_a = op_geom.verts[op_geom.tris[0]];
            vec3 op_b = op_geom.verts[op_geom.tris[1]];
            vec3 op_c = op_geom.verts[op_geom.tris[2]];

            vec3 my_n = (my_b - my_a).cross(my_c - my_b);
            vec3 op_n = (op_b - op_a).cross(op_c - op_b);

            // find y-rotation necessary to make normals opposite (in xz only)
            float my_th = atan2f(-my_n.z, my_n.x);
            float op_th = atan2f(-op_n.z, op_n.x);
            float theta = M_PI + my_th - op_th;

            // transform
            mat4 old_mview = ctx.uniform()[L_modelview];
            ctx.uniform()[L_modelview] =
                old_mview
              * mat4::translate(vec3(my_mid.x, 0.f, my_mid.y))
              * mat4::rotate_y(theta)
              * mat4::translate(-vec3(op_mid.x, 0.f, op_mid.y));

            // get everything necessary from pres here. It will be reset!
            float min_xf = pres.min_x.get(0);
            min_xf = std::min(min_xf, pres.min_x.get(1));
            min_xf = std::min(min_xf, pres.min_x.get(2));
            min_xf = std::min(min_xf, pres.min_x.get(3));

            float min_yf = pres.min_y.get(0);
            min_yf = std::min(min_yf, pres.min_y.get(1));
            min_yf = std::min(min_yf, pres.min_y.get(2));
            min_yf = std::min(min_yf, pres.min_y.get(3));

            float max_xf = pres.max_x.get(0);
            max_xf = std::max(max_xf, pres.max_x.get(1));
            max_xf = std::max(max_xf, pres.max_x.get(2));
            max_xf = std::max(max_xf, pres.max_x.get(3));

            float max_yf = pres.max_y.get(0);
            max_yf = std::max(max_yf, pres.max_y.get(1));
            max_yf = std::max(max_yf, pres.max_y.get(2));
            max_yf = std::max(max_yf, pres.max_y.get(3));

            int clip_x = int(min_xf) - 2;
            int clip_y = int(min_yf) - 2;
            int clip_w = 2 + int(max_xf) - clip_x;
            int clip_h = 2 + int(max_yf) - clip_y;

            clip_x = clamp(0, ctx.width() - 1, clip_x);
            clip_y = clamp(0, ctx.height() - 1, clip_y);
            clip_w = std::min(ctx.width()  - clip_x, clip_w);
            clip_h = std::min(ctx.height() - clip_y, clip_h);

            // bind new depth buffer
            // FIXME - these buffers should be preallocated
            int depth_x = clip_x & ~0x3;
            int depth_w = clip_w;
            if (depth_w % 4 != 0)
                depth_w += 4 - depth_w % 4;

            buffer<float>* old_depth = ctx.depthbuf();
            buffer<float> depth(ctx.width(), ctx.height());
            buffer<float> small_depth(depth_x, clip_y, depth_w, clip_h, depth);
            small_depth.clear(std::numeric_limits<float>::infinity());

            // and color buffer
            buffer<Px>* old_color = ctx.colorbuf();
            buffer<Px> color(ctx.width(), ctx.height());
            buffer<Px> small_color(depth_x, clip_y, depth_w, clip_h, color);
            small_color.clear((L_color = 0xffff0000, L_portalmask = false));

            ctx.bind_buffers(&color, &depth);

            // set clip region
            int old_clip_x, old_clip_y, old_clip_w, old_clip_h;
            ctx.clip_region(old_clip_x, old_clip_y, old_clip_w, old_clip_h);
            ctx.set_clip_region(clip_x, clip_y, clip_w, clip_h);

            // provide the current portal's depth values as a texture
            buffer<float>* old_portz = ctx.uniform()[L_portaldepth];
            ctx.uniform()[L_portaldepth] = old_depth;

            // draw behind the portal!
            other_room->draw(ctx, pres, rooms, other_port_i, portal_shader);

            // copy back to original color buffer, minding the mask
            int y_bound = clip_y + clip_h;
            int x_bound = depth_x + depth_w;
            for (int y = clip_y; y < y_bound; y++) {
                for (int x = depth_x; x < x_bound; x += 4) {
                    simd<Px>::type& color_orig = old_color->at_quad(x, y);
                    mask_4 ok = color_orig[L_portalmask];
                    color_orig[L_color] = simd<uint32_t>::choose(ok,
                        color.at_quad(x, y)[L_color], color_orig[L_color]);
                    color_orig[L_portalmask] = mask_4(false);
                }
            }

            // restore state
            ctx.uniform()[L_portaldepth] = old_portz;
            ctx.bind_buffers(old_color, old_depth);
            ctx.set_clip_region(old_clip_x, old_clip_y, old_clip_w, old_clip_h);
            ctx.uniform()[L_modelview] = old_mview;
        }
    }

    void draw_parts(context<Uni, VAttr, Px>& ctx) const {
        for (auto it = parts.begin(); it != parts.end(); it++)
            (*it)->draw(ctx);
    }
};

/**********************************************************************/

struct maze {
    std::unordered_map<std::string, room_models*> rooms_by_name;

    ~maze() {
        for (auto it = rooms_by_name.begin(); it != rooms_by_name.end(); it++)
            delete it->second;
    }

    bool load(const char* initial_room_path, const char** err) {
        *err = "";
        std::queue<std::string> q;
        q.push(std::string(initial_room_path));
        while (!q.empty()) {
            std::string next = q.front();
            q.pop();
            auto iter = rooms_by_name.find(next);
            if (iter != rooms_by_name.end())
                continue; // already have this one

            room_models* room = new room_models;
            if (!room->load(next.c_str())) {
                *err = room->error;
                auto it = rooms_by_name.begin();
                for (; it != rooms_by_name.end(); it++)
                    delete it->second;
                rooms_by_name.clear();
                return false;
            }

            rooms_by_name.insert(std::make_pair(next, room));
            auto r_it = room->portal_rooms.begin();
            for (; r_it != room->portal_rooms.end(); r_it++)
                q.push(*r_it);
        }

        return true;
    }
};

/**********************************************************************/

struct camera {
    vec3 pos, dir;

    camera(const vec3& pos, const vec3& dir)
        : pos(pos), dir(dir.normalize()) { }

    // get the rightward direction, as a unit vector
    vec3 right()   const { return dir.cross(vec3::y_basis()).normalize(); }

    // get the (horizontally) forward direction, as a unit vector
    vec3 forward() const { return vec3(dir.x, 0.f, dir.z).normalize(); }

    void look(float theta_left_right, float theta_up_down) {
        vec3 right = this->right();
        mat4 rot_y = mat4::rotate_y(theta_left_right);
        mat4 rot_r = mat4::rotate_axis(theta_up_down, right);
        vec4 d     = rot_y * dir.v4();
        vec3 d2    = (rot_r * d).xyz();
        // throw away up/down rotations that bring
        // the view direction too close to vertical
        bool bad = fabs(1.f - fabs(d2.dot(vec3::y_basis()))) < 0.05;
        dir = (bad ? d.xyz() : d2).normalize();
    }

    mat4 view_transform() {
        // I would construct view-{right,up,forward} in model space,
        // and invert that matrix, but I don't want to have to invert.
        // So assume that up is positive y.
        mat4 rot_y = mat4::rotate_y(M_PI_2 - atan2f(-dir.z, dir.x));
        vec3 dir2  = (rot_y * dir.v4()).xyz();
        mat4 rot_x = mat4::rotate_x(-atan2(dir2.y, -dir2.z));
        mat4 trans = mat4::translate(-pos);
        return rot_x * rot_y * trans;
    }
};

/**********************************************************************/

#define DEFAULT_NTHREADS      1
#define DEFAULT_SCREEN_FLAGS  (SDL_HWSURFACE)
#define DEFAULT_SCREEN_WIDTH  640
#define DEFAULT_SCREEN_HEIGHT 400
#define DEFAULT_FONT_FILE     "data/font.ttf"
#define DEFAULT_ROOM_PATH     "data/rooms/room_a"

struct options {
    int nthreads;
    int screen_flags;
    int screen_width;
    int screen_height;
    const char* font_file;
    const char* room_path;
};

options parse_options(int argc, char* argv[]);

// shared parts of pixel shader
inline uint_4 convert_colors(vec3_4 c);

int main(int argc, char *argv[]) {
    struct options opt = parse_options(argc, argv);

    // initialize px
    px_init();

    // initialize sdl
    const char* errmsg;
    if (!sdlutil_init(&errmsg)) {
        fprintf(stderr, "error: %s\n", errmsg);
        exit(1);
    }

    // open video
    unique_surface_ptr screen(SDL_SetVideoMode(
        opt.screen_width, opt.screen_height, 32, opt.screen_flags));
    if (screen == NULL) {
        fprintf(stderr, "error: SDL_SetVideoMode: %s\n", SDL_GetError());
        exit(1);
    }

    if (screen->format->BytesPerPixel != 4) {
        fprintf(stderr, "fatal: not 4 bytes per pixel!\n");
        exit(1);
    }

    SDL_WM_GrabInput(SDL_GRAB_ON);
    SDL_ShowCursor(SDL_DISABLE);

    // load a font
#ifndef NO_SDL_TTF
    unique_font_ptr font(TTF_OpenFont(opt.font_file, 24));
    if (font == NULL) {
        fprintf(stderr, "error: TTF_OpenFont: %s\n", TTF_GetError());
        exit(1);
    }

    TTF_SetFontStyle(font.get(), TTF_STYLE_BOLD);
#else
    uint32_t last_fps_print = 0;
#endif

    // load maze
    maze the_maze;
    if (!the_maze.load(opt.room_path, &errmsg)) {
        fprintf(stderr, "error: %s\n", errmsg);
        exit(1);
    }

    std::string cur_room(opt.room_path);

    // Set up renderer
    buffer<SDL_Px> sdl_buf(
        opt.screen_width, opt.screen_height,
        (simd<SDL_Px>::type*)screen->pixels);

    buffer<Px> target(opt.screen_width, opt.screen_height);
    buffer<float> top_portaldepth(opt.screen_width, opt.screen_height);
    top_portaldepth.clear(0.f);

    Uni initial_uni = (L_projection = mat4::identity(),
                       L_modelview  = mat4::identity(),
                       L_texture    = (buffer<vec3>*)NULL,
                       L_portaldepth = &top_portaldepth);

    buffer<float> zbuffer(opt.screen_width, opt.screen_height);

    worker_pool wpool(opt.nthreads);
    if (!wpool.init()) {
        fprintf(stderr, "fatal: failed to initialize threads!\n");
        exit(1);
    }

    context<Uni, VAttr, Px> raster(wpool, initial_uni);

    auto default_shader = PX_MAKE_SHADER_OBJ(raster,
    ([](const Uni& uni, const vec3& v, const VAttr& attr) {
        return (L_pos     = uni[L_projection] * uni[L_modelview] * v.p4(),
                L_qlinear = (L_texcoord = attr[L_texcoord]));
    }),
    ([](const Uni& uni, const varying_t& vary, int_4 x, int_4 y,
        float_4& z, simd<Px>::type& color) {
        float_4 port_z = uni[L_portaldepth]->at(x, y);
        // leave L_portalmask as is
        vec3_4 c = uni[L_texture]->sample(vary[L_qlinear][L_texcoord]);
        uint_4 c_orig = color[L_color];
        color[L_color] = simd<uint32_t>::choose(z >= port_z,
            convert_colors(c), c_orig);
    }));

    portal_results portal_res;
    auto portal_shader = PX_MAKE_SHADER_OBJ(raster,
    ([](const Uni& uni, const vec3& v, const VAttr& attr) {
        return (L_pos = uni[L_projection] * uni[L_modelview] * v.p4());
    }),
    ([&portal_res](const Uni& uni, const varying_t& vary, int_4 x, int_4 y,
        float_4& z, simd<Px>::type& color) {
        float_4 port_z = uni[L_portaldepth]->at(x, y);
        mask_4 behind_port = z >= port_z;
        // red, and set mask
        mask_4 portmask_orig = color[L_portalmask];
        color[L_portalmask] = simd<bool>::choose(behind_port,
            mask_4(true), portmask_orig);
        // update portal_res
        if (behind_port.none())
            return;
        float_4 xf(x), yf(y);
        portal_res.hit = true;
        portal_res.min_x.min_d(xf);
        portal_res.min_y.min_d(yf);
        portal_res.max_x.max_d(xf);
        portal_res.max_y.max_d(yf);
    }));

    raster.bind_shader(default_shader);

    // camera state
    camera cam(vec3(.25f, 1.5f, -.25f), vec3(1.f, 0.f, 1.f));

    enum walk_dir_t { WALK_LEFT_RIGHT = 0, WALK_FWD_BACK = 1 };
    int walk[2] = { 0, 0 };

    // want to keep a framerate
    uint32_t last_frame = 0;
    framerate_meter fpsmeter(250);

    for (;;) {
    // keep track of time
    uint32_t now = fpsmeter.tick(SDL_GetTicks());
    uint32_t frame_time = now - last_frame;
    last_frame = now;

    // handle events
    SDL_Event ev;
    bool quit = false;
    while (SDL_PollEvent(&ev)) {
        switch (ev.type) {
        case SDL_MOUSEMOTION: {
            float halfdeg = M_PI_2 / 180.f;
            cam.look(-halfdeg * ev.motion.xrel, -halfdeg * ev.motion.yrel);
            break;
        }

        case SDL_KEYDOWN:
            switch (ev.key.keysym.sym) {
            case SDLK_ESCAPE:
                quit = true;
                break;
            case SDLK_c:
                raster.enable_cull(!raster.cull_enabled());
                break;

            case FORWARD_KEY: walk[WALK_FWD_BACK  ]++; break;
            case LEFT_KEY:    walk[WALK_LEFT_RIGHT]--; break;
            case BACK_KEY:    walk[WALK_FWD_BACK  ]--; break;
            case RIGHT_KEY:   walk[WALK_LEFT_RIGHT]++; break;

            default:
                ; // ignore keydown
            }
            break;

        case SDL_KEYUP:
            switch (ev.key.keysym.sym) {
            // opposite the above
            case FORWARD_KEY: walk[WALK_FWD_BACK  ]--; break;
            case LEFT_KEY:    walk[WALK_LEFT_RIGHT]++; break;
            case BACK_KEY:    walk[WALK_FWD_BACK  ]++; break;
            case RIGHT_KEY:   walk[WALK_LEFT_RIGHT]--; break;
            default:
                ; // ignore keyup
            }

        default:
            ; // ignore event
        }
    }

    if (quit)
        break;

    // walk
    vec3 walk_vectors[2] = { cam.right(), cam.forward() };
    if (walk[WALK_LEFT_RIGHT] != 0 || walk[WALK_FWD_BACK] != 0) {
        vec3 walk_dir = vec3::zero();
        for (int i = 0; i < 2; i++)
            walk_dir += walk[i] * walk_vectors[i];

        cam.pos += frame_time / 600.f * walk_dir.normalize();
    }

    // Get access to screen pixels
    if (SDL_MUSTLOCK(screen.get())) {
        if (SDL_LockSurface(screen.get()) < 0) {
            fprintf(stderr, "fatal: unable to lock screen!\n");
            exit(1);
        }
    }

    // projection transformation
    raster.uniform()[L_projection] = mat4::perspective(
        .25f * M_PI, opt.screen_width, opt.screen_height, .01f, 256.f);

    raster.set_near_distance(.01f);

    // modelview transformation - view part
    raster.uniform()[L_modelview] = cam.view_transform();

    // draw
    raster.bind_buffers(&target, &zbuffer);
    raster.set_cull_face(CULL_BACK);
    target.clear((L_color = 0x0000007f, L_portalmask = false));
    zbuffer.clear(std::numeric_limits<float>::infinity());

    the_maze.rooms_by_name.at(cur_room)->draw(
        raster, portal_res, the_maze.rooms_by_name, -1, portal_shader);

    //room.draw(raster);
    //hall.draw(raster);

    //buffer<Px> sub_buf(12, 12, 100, 75, target);
    //sub_buf.clear(0x00ff00); // green

    //raster.draw_splits(0xff0000);

    // copy from target to sdl_buf
    for (int y = 0; y < target.height(); y++) {
        for (int x = 0; x < target.width(); x += 4)
            sdl_buf.at_quad(x, y) = target.at_quad(x, y)[L_color];
    }

    // Done rendering, unlock the screen.
    if (SDL_MUSTLOCK(screen.get()))
        SDL_UnlockSurface(screen.get());

    // write FPS to the screen
    int fr = fpsmeter.framerate() + .5f;
    char fr_msg[32];
    snprintf(fr_msg, sizeof (fr_msg), "fps:%d", fr);

#ifndef NO_SDL_TTF
    SDL_Color black = { 0, 0, 0 };
    SDL_Color white_palette[] = { { 0, 0, 0 }, { 0xff, 0xff, 0xff } };
    unique_surface_ptr msg_surface(TTF_RenderText_Solid(
        font.get(), fr_msg, black));
    if (msg_surface == NULL) {
        // non-fatal
        fprintf(stderr, "error: TTF_RenderText_Solid: %s\n", TTF_GetError());
    } else {
        SDL_Rect dst_rect = { .x = 12, .y = 12 }; // w, h ignored
        SDL_BlitSurface(msg_surface.get(), NULL, screen.get(), &dst_rect);
        SDL_SetPalette(msg_surface.get(), SDL_LOGPAL, white_palette, 0, 2);
        dst_rect.x = 10;
        dst_rect.y = 10;
        SDL_BlitSurface(msg_surface.get(), NULL, screen.get(), &dst_rect);
    }
#else
    if (now - last_fps_print >= 250) {
        last_fps_print = now;
        printf("%s\n", fr_msg);
    }
#endif

    // Update the entire screen
    SDL_UpdateRect(screen.get(), 0, 0, opt.screen_width, opt.screen_height);

    } // SDL_GetTicks() loop

    // Done
    return 0;
}

inline uint_4 convert_colors(vec3_4 c) {
    float_4 zero(0.f), one(1.f);

    c.x.clamp_d(zero, one);
    c.y.clamp_d(zero, one);
    c.z.clamp_d(zero, one);

    c *= float_4(255.f);
    uint_4 rgba(0xff);
    rgba <<= 8;
    rgba  |= uint_4(c.x);
    rgba <<= 8;
    rgba  |= uint_4(c.y);
    rgba <<= 8;
    rgba  |= uint_4(c.z);

    return rgba;
}

options parse_options(int argc, char *argv[]) {
    // defaults
    options opt;
    opt.nthreads      = DEFAULT_NTHREADS;
    opt.screen_flags  = DEFAULT_SCREEN_FLAGS;
    opt.screen_width  = DEFAULT_SCREEN_WIDTH;
    opt.screen_height = DEFAULT_SCREEN_HEIGHT;
    opt.font_file     = DEFAULT_FONT_FILE;
    opt.room_path     = DEFAULT_ROOM_PATH;

    // override with switches
    for (int i = 1; i < argc; i++) {
        if (strcmp(argv[i], "-n") == 0) {
            if (++i < argc) {
                char* arg = argv[i];
                opt.nthreads = strtol(arg, &arg, 0);
                if (opt.nthreads < 1 || opt.nthreads > 32 || *arg != '\0') {
                    fprintf(stderr, "error: invalid number of threads %s "
                            "(should be an integer between 1 and 32)\n",
                            argv[i]);
                    exit(2);
                }
            } else {
                fprintf(stderr, "error: -n requires an argument\n");
                exit(2);
            }
        } else if (strcmp(argv[i], "-fs") == 0) {
            opt.screen_flags |= SDL_FULLSCREEN;
        } else if (strcmp(argv[i], "-dim") == 0) {
            if (++i < argc) {
                char* arg = argv[i];
                opt.screen_width = strtol(arg, &arg, 0);
                bool ok = true;
                if (*arg != 'x') {
                    ok = false;
                } else {
                    arg++;
                    opt.screen_height = strtol(arg, &arg, 0);
                    ok = *arg == '\0';
                }

                if (!ok) {
                    fprintf(stderr, "error: invalid dimension %s "
                            "(should be wxh, e.g. 800x600)\n", argv[i]);
                    exit(2);
                } else if (opt.screen_width % 4 != 0) {
                    fprintf(stderr, "error: width %d not a multiple of 4\n",
                            opt.screen_width);
                    exit(2);
                } else if (opt.screen_width < 0 || opt.screen_height < 0) {
                    fprintf(stderr, "error: negative dimension %s\n", argv[i]);
                    exit(2);
                }
            } else {
                fprintf(stderr, "error: -dim requires an argument\n");
                exit(2);
            }
        } else if (strcmp(argv[i], "-fn") == 0) {
            if (++i < argc) {
                opt.font_file = argv[i];
            } else {
                fprintf(stderr, "error: -fn requires an argument\n");
                exit(2);
            }
        } else if (strcmp(argv[i], "-room") == 0) {
            if (++i < argc) {
                opt.room_path = argv[i];
            } else {
                fprintf(stderr, "error: -room requires an argument\n");
                exit(2);
            }
        } else {
            fprintf(stderr, "error: unrecognized argument %s\n", argv[i]);
            exit(2);
        }
    }

    return opt;
}
