#include "mesh.hh"

#include <vector>
#include <unordered_map>

#include <cstdio>
#include <cstring>
#include <cerrno>

bool mesh::load(const char* filename) {
    error[0] = '\0';

    if (verts != NULL) {
        snprintf(error, sizeof (error), "mesh already loaded");
        return false;
    }

    obj_info obj;
    if (!read_obj(filename, obj))
        return false;

    // duplicate vertices for each uv/normal pair
    std::unordered_map<ind3, uint16_t, ind3_hash> map;
    std::vector<px::vec3> v, vn;
    std::vector<px::vec2> vt;
    std::vector<uint16_t> f;

    for (auto iter = obj.f.begin(); iter != obj.f.end(); ++iter) {
        for (int i = 0; i < 3; i++) {
            ind3 i3 = iter->inds[i];
            auto entry = map.find(i3);
            int  j;
            if (entry != map.end()) {
                j = entry->second;
            } else {
                j = v.size();
                if (i3.vi >= obj.v.size()) {
                    snprintf(error, sizeof (error),
                             "vertex index %d out of range", i3.vi);
                    return false;
                } else if (i3.vti >= obj.vt.size()) {
                    snprintf(error, sizeof (error),
                             "texture coordinate index %d out of range", i3.vti);
                    return false;
                } else if (i3.vni >= obj.vn.size()) {
                    snprintf(error, sizeof (error),
                             "vertex normal index %d out of range", i3.vni);
                    return false;
                }
                v.push_back(obj.v[i3.vi]);
                vt.push_back(obj.vt[i3.vti]);
                vn.push_back(obj.vn[i3.vni]);
                map.insert(std::pair<ind3, uint16_t>(i3, j));
            }

            f.push_back(j);
        }
    }

    // set stuff
    nverts = v.size();
    ntris  = f.size() / 3;

    verts     = new px::vec3[nverts];
    texcoords = new px::vec2[nverts];
    normals   = new px::vec3[nverts];
    for (int i = 0; i < nverts; i++) {
        verts[i]     = v[i];
        texcoords[i] = vt[i];
        normals[i]   = vn[i];
    }

    tris = new uint16_t[ntris * 3];
    for (int i = 0; i < ntris * 3; i++)
        tris[i] = f[i];

    return true;
}

void mesh::unload() {
    error[0] = '\0';
    if (verts == NULL)
        return;

    delete[] verts;
    delete[] normals;
    delete[] texcoords;
    delete[] tris;
    verts     = NULL;
    normals   = NULL;
    texcoords = NULL;
    nverts    = 0;
    ntris     = 0;
}

bool mesh::read_obj(const char* filename, obj_info& obj) {
    FILE* fp = fopen(filename, "r");
    if (fp == NULL) {
        snprintf(error, sizeof (error), "%s: %s",
                 filename, strerror(errno));
        return false;
    }

    bool result = read_obj(filename, (void*)fp, obj);
    fclose(fp);
    return result;
}

bool mesh::read_obj(const char* filename, void* fp_void, obj_info& obj) {
    FILE* fp = (FILE*)fp_void;
    const int MAX_NTOK = 5;

    char  line[128];
    char* tokens[MAX_NTOK];
    int   line_no = 0;
    while (fgets(line, sizeof (line), fp) != NULL) {
        line_no++;
        int line_len = strlen(line);
        if (line_len == sizeof(line) - 1) {
            snprintf(error, sizeof (error),
                     "%s:%d: line length limit exceeded",
                     filename, line_no);
            return false;
        }

        // remove trailing newline if any
        if (line_len >= 1 && line[line_len - 1] == '\n')
            line[--line_len] = '\0';

        // split on whitespace
        int ntok = 0;
        if (!split_ws(line, tokens, ntok, MAX_NTOK)) {
            snprintf(error, sizeof (error),
                     "%s:%d: too many whitespace-separated tokens",
                     filename, line_no);
            return false;
        }

        if (ntok == 0)
            continue; // comment or empty line

        // read data point
        if (!read_line(filename, line_no, tokens, ntok, obj))
            return false;
    }

    return true;
}

bool mesh::read_line(const char* filename, int line_no,
                      char* tok[], int ntok, obj_info& obj) {
    const char* ignore[] = { "o", "g", "s", NULL };
    for (int i = 0; ignore[i] != NULL; i++) {
        if (strcmp(tok[0], ignore[i]) == 0)
            return true;
    }

    if (strcmp(tok[0], "v") == 0) {
        if (ntok != 4) {
            snprintf(error, sizeof (error),
                     "%s:%d: wrong number of arguments to v (%d)",
                     filename, line_no, ntok - 1);
            return false;
        }
        px::vec3 v;
        if (!read_vec3(v, tok[1], tok[2], tok[3])) return false;
        obj.v.push_back(v);
    } else if (strcmp(tok[0], "vn") == 0) {
        if (ntok != 4) {
            snprintf(error, sizeof (error),
                     "%s:%d: wrong number of arguments to vn (%d)",
                     filename, line_no, ntok - 1);
            return false;
        }
        px::vec3 vn;
        if (!read_vec3(vn, tok[1], tok[2], tok[3])) return false;
        obj.vn.push_back(vn.normalize());
    } else if (strcmp(tok[0], "vt") == 0) {
        if (ntok != 3) {
            snprintf(error, sizeof (error),
                     "%s:%d: wrong number of arguments to vt (%d)",
                     filename, line_no, ntok - 1);
            return false;
        }
        px::vec3 vt;
        if (!read_vec3(vt, tok[1], tok[2], "0")) return false;
        obj.vt.push_back(px::vec2(vt.x, vt.y));
    } else if (strcmp(tok[0], "f") == 0) {
        if (ntok != 4 && ntok != 5) {
            snprintf(error, sizeof (error),
                     "%s:%d: wrong number of arguments to f (%d)"
                     "; not a triangle or quad",
                     filename, line_no, ntok - 1);
            return false;
        }
        tri_info tri;
        ind3 extra = { 0, 0, 0 }; // for quads
        bool err = false;
        err = err || !read_ind3(tri.inds[0], tok[1]);
        err = err || !read_ind3(tri.inds[1], tok[2]);
        err = err || !read_ind3(tri.inds[2], tok[3]);
        if (ntok == 5)
            err = err || !read_ind3(extra, tok[4]);
        if (err) {
            snprintf(error, sizeof (error),
                     "%s:%d: unable to parse indices",
                     filename, line_no);
            return false;
        }
        obj.f.push_back(tri);
        if (ntok == 5) {
            tri.inds[1] = tri.inds[2];
            tri.inds[2] = extra;
            obj.f.push_back(tri);
        }
    } else {
        snprintf(error, sizeof (error),
                 "%s:%d: unrecognized definition %s",
                 filename, line_no, tok[0]);
        return false;
    }

    return true;
}

bool mesh::read_vec3(
    px::vec3& out, const char* x, const char* y, const char* z) {
    char* end;
    out.x = strtof(x, &end);
    if (*end != '\0')
        return false;

    out.y = strtof(y, &end);
    if (*end != '\0')
        return false;

    out.z = strtof(z, &end);
    if (*end != '\0')
        return false;

    return true;
}

bool mesh::read_ind3(ind3& out, char* s) {
    out.vi = strtol(s, &s, 0) - 1;
    if (*s++ != '/')
        return false;

    out.vti = strtol(s, &s, 0) - 1;
    if (*s++ != '/')
        return false;

    out.vni = strtol(s, &s, 0) - 1;
    return *s == '\0';
}

bool mesh::split_ws(char* line, char* tokens[], int& ntok, int max_ntok) {
    int i = 0;
    while (isspace(line[i]))
        i++;

    ntok = 0;
    if (line[i] == '#' || line[i] == '\0')
        return true; // empty line (whitespace and/or comment only)

    tokens[ntok++] = &line[i];

    for (;;) {
        while (line[i] != '\0' && !isspace(line[i]))
            i++;

        if (line[i] == '\0')
            break;

        line[i++] = '\0';
        while (isspace(line[i]))
            i++;

        if (line[i] == '\0')
            break;

        if (ntok >= max_ntok)
            return false;

        tokens[ntok++] = &line[i];
    }

    return true;
}
