#ifndef _PX_BUFFER_HH_
#define _PX_BUFFER_HH_

#include <cassert>
#include <cfenv>

// requires px/linalg_simd.hh

namespace px {

template<class U>
class buffer {
    int const w_, h_;

    typedef typename simd<U>::type simd_t;
    int     const pitch_;  // number of simd_t in a row.
    simd_t* const ptr_;
    bool    const free_;

public:
    buffer(int w, int h)
        : w_(w),  h_(h)
        , pitch_(w / 4 + (w % 4 == 0 ? 0 : 1))
        , ptr_(new simd_t[pitch_ * h])
        , free_(true) { }

    buffer(int w, int h, simd_t* ptr)
        : w_(w),  h_(h)
        , pitch_(w / 4 + (w % 4 == 0 ? 0 : 1))
        , ptr_(ptr)
        , free_(false) { }

    // "window" into another buffer
    buffer(int x, int y, int w, int h, buffer<U>& buf)
        : w_(w), h_(h)
        , pitch_(buf.pitch_)
        , ptr_(&buf.ptr_[y * pitch_ + x / 4])
        , free_(false) {
        assert(x % 4 == 0);
        assert(w % 4 == 0);
        assert(x + w_ <= buf.w_);
        assert(y + h_ <= buf.h_);
    }

    ~buffer() {
        if (free_)
            delete ptr_;
    }

    int width()  const { return w_; }
    int height() const { return h_; }
    int pitch()  const { return pitch_; }

    class elt_ref_base {
    protected:
        simd_t* p_;
        int     i_;
    public:
        elt_ref_base(simd_t* p, int i) : p_(p), i_(i) { }
        U get() const { return simd<U>::get(*p_, i_); }
        operator U() const { return get(); }
    };

    class const_elt_ref : public elt_ref_base {
        const_elt_ref(const simd_t* p, int i)
            : elt_ref_base(const_cast<simd_t*>(p), i) { }
    };

    struct elt_ref : public elt_ref_base {
        elt_ref(simd_t* p, int i) : elt_ref_base(p, i) { }
        elt_ref& operator=(const U& x) {
            simd<U>::set(*this->p_, this->i_, x);
            return *this;
        }
    };

    const_elt_ref at(int s, int t) const {
        check_int(s, t);
        return const_elt_ref(&ptr_[pitch_ * t + (s >> 2)], s & 0x3);
    }

    elt_ref at(int s, int t) {
        check_int(s, t);
        return elt_ref(&ptr_[pitch_ * t + (s >> 2)], s & 0x3);
    }

    const simd_t& at_quad(int s, int t) const {
        assert((s & 0x3) == 0);
        return ptr_[pitch_ * t + (s >> 2)];
    }

    simd_t& at_quad(int s, int t) {
        assert((s & 0x3) == 0);
        return ptr_[pitch_ * t + (s >> 2)];
    }

    const_elt_ref sample(float s, float t) const {
        return at(scale(s, float(w_)), scale(t, float(h_)));
    }

    const_elt_ref sample(const vec2& uv) const { return sample(uv.x, uv.y); }

    elt_ref sample(float s, float t) {
        return at(scale(s, float(w_)), scale(t, float(h_)));
    }

    void clear(const U& u) {
        simd_t u4 = simd<U>::make(u, u, u, u);
        for (int y = 0; y < h_; y++) {
            simd_t* start = &ptr_[y * pitch_];
            simd_t* end   = start + (w_ >> 2); // not w_! not pitch_!
            for (simd_t* p = start; p != end; p++)
                *p = u4;
        }
    }

    simd_t at(const int_4& s, const int_4& t) const {
        check_int4(s, t);
        int_4 is = int_4(pitch_) * t + (s >> 2);
        int_4 js = s & int_4(0x3);
        return simd<U>::make(
            simd<U>::get(ptr_[is.at(0)], js.at(0)),
            simd<U>::get(ptr_[is.at(1)], js.at(1)),
            simd<U>::get(ptr_[is.at(2)], js.at(2)),
            simd<U>::get(ptr_[is.at(3)], js.at(3)));
    }

    simd_t sample(const float_4& s, const float_4& t) const {
        return at(scale(s, float_4(w_)), scale(t, float_4(h_)));
    }

    simd_t sample(const vec2_4& uv) const { return sample(uv.x, uv.y); }

private:
    void check_int4(const int_4& s, const int_4& t) const {
        assert((s >= int_4(0)).all() && (s < int_4(w_)).all());
        assert((t >= int_4(0)).all() && (t < int_4(h_)).all());
    }

    void check_int(int s, int t) const {
        assert(s >= 0 && s < w_);
        assert(t >= 0 && t < h_);
    }

    static float wrap(float s) {
        return s - floorf(s);
    }

    static int scale(float s, float w) {
        return (int)clamp(0.f, w-.5f, w * wrap(s));
    }

    static float_4 wrap(const float_4& s) {
        int_4   floor_i(s);
        float_4 floor = float_4(floor_i);
        return s - floor;
    }

    static int_4 scale(const float_4& s, const float_4& w) {
        return int_4((w * wrap(s)).clamp(float_4(0.f), w-float_4(.5f)));
    }
};

}      // namespace px
#endif // _PX_BUFFER_HH_
