#ifndef _PX_LINALG_HH_
#define _PX_LINALG_HH_

#include <algorithm>
#include <cstddef>

namespace px {

template<typename U>
inline const U& clamp(const U& a, const U& b, const U& x) {
    return std::max(a, std::min(b, x));
}

struct vec2;
struct vec3;
struct vec4;
inline vec2 operator*(float a, const vec2& v);
inline vec3 operator*(float a, const vec3& v);
inline vec4 operator*(float a, const vec4& v);

struct vec4 {
    union {
        struct { float x, y, z, w; };
        float coords[4];
    };

    // construction
    vec4() { } // avoid
    vec4(float x, float y, float z, float w) : x(x), y(y), z(z), w(w) { }

    static vec4 zero()    { return vec4(0.f, 0.f, 0.f, 0.f); }
    static vec4 x_basis() { return vec4(1.f, 0.f, 0.f, 0.f); }
    static vec4 y_basis() { return vec4(0.f, 1.f, 0.f, 0.f); }
    static vec4 z_basis() { return vec4(0.f, 0.f, 1.f, 0.f); }
    static vec4 w_basis() { return vec4(0.f, 0.f, 0.f, 1.f); }

    // assignment
    vec4& operator=(const vec4& v) {
        x = v.x;
        y = v.y;
        z = v.z;
        w = v.w;
        return *this;
    }

    // component-wise arithmetic
    vec4 operator-() const { return vec4(-x, -y, -z, -w); }
    vec4 operator+(const vec4& v) const { return vec4(x+v.x, y+v.y, z+v.z, w+v.w); }
    vec4 operator-(const vec4& v) const { return vec4(x-v.x, y-v.y, z-v.z, w-v.w); }
    vec4 operator*(const vec4& v) const { return vec4(x*v.x, y*v.y, z*v.z, w*v.w); }
    vec4 operator/(const vec4& v) const { return vec4(x/v.x, y/v.y, z/v.z, w/v.w); }

    // destructive component-wise arithmetic
    vec4& operator+=(const vec4& v) { x+=v.x; y+=v.y; z+=v.z; w+=v.w; return *this; }
    vec4& operator-=(const vec4& v) { x-=v.x; y-=v.y; z-=v.z; w-=v.w; return *this; }
    vec4& operator*=(const vec4& v) { x*=v.x; y*=v.y; z*=v.z; w*=v.w; return *this; }
    vec4& operator/=(const vec4& v) { x/=v.x; y/=v.y; z/=v.z; w/=v.w; return *this; }

    // scalar arithmetic
    vec4  operator*(float a) const { return vec4(a*x, a*y, a*z, a*w); }
    vec4  operator/(float a) const { return vec4(x/a, y/a, z/a, w/a); }
    vec4& operator*=(float a) { x*=a; y*=a; z*=a; w*=a; return *this; }
    vec4& operator/=(float a) { x/=a; y/=a; z/=a; w/=a; return *this; }

    // dot product
    float dot(const vec4& v) const { return x*v.x + y*v.y + z*v.z + w*v.w; }

    // norm
    float norm_sqr() const { return dot(*this); }
    float norm()     const { return sqrtf(norm_sqr()); }
    float rnorm()    const { return 1.f / norm(); }

    // normalize
    vec4  normalize() const { return rnorm() * (*this); }
    vec4& normalize_d() { return *this *= rnorm(); }

    // dropping components / swizzling
    inline vec3 xyz() const;

private:
    static void check_offsets() {
        const int xoffs = 0 * sizeof (float);
        const int yoffs = 1 * sizeof (float);
        const int zoffs = 2 * sizeof (float);
        const int woffs = 3 * sizeof (float);
        static_assert(offsetof(vec4, x) == xoffs, "unexpected offset");
        static_assert(offsetof(vec4, y) == yoffs, "unexpected offset");
        static_assert(offsetof(vec4, z) == zoffs, "unexpected offset");
        static_assert(offsetof(vec4, w) == woffs, "unexpected offset");
    }
};

struct vec3 {
   union {
       struct { float x, y, z; };
       float coords[4];
   };

    // construction
    vec3() { } // avoid
    vec3(float x, float y, float z) : x(x), y(y), z(z) { }

    static vec3 zero()    { return vec3(0.f, 0.f, 0.f); }
    static vec3 x_basis() { return vec3(1.f, 0.f, 0.f); }
    static vec3 y_basis() { return vec3(0.f, 1.f, 0.f); }
    static vec3 z_basis() { return vec3(0.f, 0.f, 1.f); }

    // assignment
    vec3& operator=(const vec3& v) {
        x = v.x;
        y = v.y;
        z = v.z;
        return *this;
    }

    // component-wise arithmetic
    vec3 operator-() const { return vec3(-x, -y, -z); }
    vec3 operator+(const vec3& v) const { return vec3(x+v.x, y+v.y, z+v.z); }
    vec3 operator-(const vec3& v) const { return vec3(x-v.x, y-v.y, z-v.z); }
    vec3 operator*(const vec3& v) const { return vec3(x*v.x, y*v.y, z*v.z); }
    vec3 operator/(const vec3& v) const { return vec3(x/v.x, y/v.y, z/v.z); }

    // destructive component-wise arithmetic
    vec3& operator+=(const vec3& v) { x+=v.x; y+=v.y; z+=v.z; return *this; }
    vec3& operator-=(const vec3& v) { x-=v.x; y-=v.y; z-=v.z; return *this; }
    vec3& operator*=(const vec3& v) { x*=v.x; y*=v.y; z*=v.z; return *this; }
    vec3& operator/=(const vec3& v) { x/=v.x; y/=v.y; z/=v.z; return *this; }

    // scalar arithmetic
    vec3  operator*(float a) const { return vec3(a*x, a*y, a*z); }
    vec3  operator/(float a) const { return vec3(x/a, y/a, z/a); }
    vec3& operator*=(float a) { x*=a; y*=a; z*=a; return *this; }
    vec3& operator/=(float a) { x/=a; y/=a; z/=a; return *this; }

    // dot product
    float dot(const vec3& v) const { return x*v.x + y*v.y + z*v.z; }

    // cross product
    vec3 cross(const vec3& v) const {
        return vec3(y*v.z - z*v.y, z*v.x - x*v.z, x*v.y - y*v.x);
    }

    // norm
    float norm_sqr() const { return dot(*this); }
    float norm()     const { return sqrtf(norm_sqr()); }
    float rnorm()    const { return 1.f / norm(); }

    // normalize
    vec3  normalize() const { return rnorm() * (*this); }
    vec3& normalize_d() { return *this *= rnorm(); }

    // vec4 forms
    vec4 v4() const { return vec4(x, y, z, 0.f); }
    vec4 p4() const { return vec4(x, y, z, 1.f); }

private:
    static void check_offsets() {
        const int xoffs = 0 * sizeof (float);
        const int yoffs = 1 * sizeof (float);
        const int zoffs = 2 * sizeof (float);
        static_assert(offsetof(vec4, x) == xoffs, "unexpected offset");
        static_assert(offsetof(vec4, y) == yoffs, "unexpected offset");
        static_assert(offsetof(vec4, z) == zoffs, "unexpected offset");
    }
};

struct vec2 {
    union {
        struct { float x, y; };
        float coords[2];
    };

    // construction
    vec2() { } // avoid
    vec2(float x, float y) : x(x), y(y) { }

    static vec2 zero()    { return vec2(0.f, 0.f); }
    static vec2 x_basis() { return vec2(1.f, 0.f); }
    static vec2 y_basis() { return vec2(0.f, 1.f); }

    // assignment
    vec2& operator=(const vec2& v) {
        x = v.x;
        y = v.y;
        return *this;
    }

    // component-wise arithmetic
    vec2 operator-() const { return vec2(-x, -y); }
    vec2 operator+(const vec2& v) const { return vec2(x+v.x, y+v.y); }
    vec2 operator-(const vec2& v) const { return vec2(x-v.x, y-v.y); }
    vec2 operator*(const vec2& v) const { return vec2(x*v.x, y*v.y); }
    vec2 operator/(const vec2& v) const { return vec2(x/v.x, y/v.y); }

    // destructive component-wise arithmetic
    vec2& operator+=(const vec2& v) { x+=v.x; y+=v.y; return *this; }
    vec2& operator-=(const vec2& v) { x-=v.x; y-=v.y; return *this; }
    vec2& operator*=(const vec2& v) { x*=v.x; y*=v.y; return *this; }
    vec2& operator/=(const vec2& v) { x/=v.x; y/=v.y; return *this; }

    // scalar arithmetic
    vec2  operator*(float a) const { return vec2(a*x, a*y); }
    vec2  operator/(float a) const { return vec2(x/a, y/a); }
    vec2& operator*=(float a) { x*=a; y*=a; return *this; }
    vec2& operator/=(float a) { x/=a; y/=a; return *this; }

    // dot product
    float dot(const vec2& v) const { return x*v.x + y*v.y; }

    // norm
    float norm_sqr() const { return dot(*this); }
    float norm()     const { return sqrtf(norm_sqr()); }
    float rnorm()    const { return 1.f / norm(); }

    // normalize
    vec2  normalize() const { return rnorm() * (*this); }
    vec2& normalize_d() { return *this *= rnorm(); }
};

inline vec2 operator*(float a, const vec2& v) { return v * a; }
inline vec3 operator*(float a, const vec3& v) { return v * a; }
inline vec4 operator*(float a, const vec4& v) { return v * a; }

inline vec3 vec4::xyz() const { return vec3(x, y, z); }

struct mat4 {
    vec4 col_[4];

    // initialization
    mat4(float a00, float a01, float a02, float a03,
         float a10, float a11, float a12, float a13,
         float a20, float a21, float a22, float a23,
         float a30, float a31, float a32, float a33)
         : col_{vec4(a00, a10, a20, a30),
                vec4(a01, a11, a21, a31),
                vec4(a02, a12, a22, a32),
                vec4(a03, a13, a23, a33)}
    { }

    static mat4 zero() {
        vec4 vzero = vec4::zero();
        return mat4(vzero, vzero, vzero, vzero);
    }

    static mat4 identity() {
        return mat4(vec4::x_basis(),
                    vec4::y_basis(),
                    vec4::z_basis(),
                    vec4::w_basis());
    }

    static mat4 with_cols(const vec4& col0, const vec4& col1,
                          const vec4& col2, const vec4& col3) {
        return mat4(col0, col1, col2, col3);
    }

    static mat4 with_rows(const vec4& row0, const vec4& row1,
                          const vec4& row2, const vec4& row3) {
        return mat4(vec4(row0.x, row1.x, row2.x, row3.x),
                    vec4(row0.y, row1.y, row2.y, row3.y),
                    vec4(row0.z, row1.z, row2.z, row3.z),
                    vec4(row0.w, row1.w, row2.w, row3.w));
    }

    // persective transformations
    // TODO ortho
    static mat4 perspective(float fov_y, int w, int h, float near, float far) {
        float cot_fov2 = 1.f / tan(fov_y * .5f);
        float r_aspect = float(h) / float(w);
        float a = r_aspect * cot_fov2;
        float b = cot_fov2;
        float c = (far + near) / (near - far);
        float d = -2.f * far * near / (far - near);
        return mat4(  a, 0.f,  0.f, 0.f,
                    0.f,   b,  0.f, 0.f,
                    0.f, 0.f,    c,   d,
                    0.f, 0.f, -1.f, 0.f);
    }

    // affine transformations
    static mat4 translate(const vec3& v) {
        return mat4(1.f, 0.f, 0.f, v.x,
                    0.f, 1.f, 0.f, v.y,
                    0.f, 0.f, 1.f, v.z,
                    0.f, 0.f, 0.f, 1.f);
    }

    static mat4 nonuniform_scale(const vec3& v) {
        return mat4(v.x, 0.f, 0.f, 0.f,
                    0.f, v.y, 0.f, 0.f,
                    0.f, 0.f, v.z, 0.f,
                    0.f, 0.f, 0.f, 1.f);
    }

    static mat4 scale(float a) {
        return mat4(  a, 0.f, 0.f, 0.f,
                    0.f,   a, 0.f, 0.f,
                    0.f, 0.f,   a, 0.f,
                    0.f, 0.f, 0.f, 1.f);
    }

    static mat4 rotate_x(float theta) {
        float cth = cosf(theta);
        float sth = sinf(theta);
        return mat4(1.f,  0.f,  0.f, 0.f,
                    0.f,  cth, -sth, 0.f,
                    0.f,  sth,  cth, 0.f,
                    0.f,  0.f,  0.f, 1.f);
    }

    static mat4 rotate_y(float theta) {
        float cth = cosf(theta);
        float sth = sinf(theta);
        return mat4(cth, 0.f, sth, 0.f,
                    0.f, 1.f, 0.f, 0.f,
                   -sth, 0.f, cth, 0.f,
                    0.f, 0.f, 0.f, 1.f);
    }

    static mat4 rotate_z(float theta) {
        float cth = cosf(theta);
        float sth = sinf(theta);
        return mat4(cth, -sth,  0.f, 0.f,
                    sth,  cth,  0.f, 0.f,
                    0.f,  0.f,  1.f, 0.f,
                    0.f,  0.f,  0.f, 1.f);
    }

    // axis should be normalized
    static mat4 rotate_axis(float theta, const vec3 axis) {
        float a   = axis.x;
        float b   = axis.y;
        float c   = axis.z;
        float cth = cosf(theta);
        float sth = sinf(theta);
        float cth_c = 1.f - cth;

        float aa  = a * a * cth_c;
        float ab  = a * b * cth_c;
        float ac  = a * c * cth_c;
        float bb  = b * b * cth_c;
        float bc  = b * c * cth_c;
        float cc  = c * c * cth_c;

        float sth_a = sth * a;
        float sth_b = sth * b;
        float sth_c = sth * c;

        return mat4(
              cth + aa, -sth_c + ab,  sth_b + ac, 0.f,
            sth_c + ab,    cth + bb, -sth_a + bc, 0.f,
           -sth_b + ac,  sth_a + bc,    cth + cc, 0.f,
                   0.f,         0.f,         0.f, 1.f
        );
    }

    // multiplication
    vec4 operator*(const vec4& v) const {
        float x, y, z, w;

        x  = v.x * col_[0].x;
        x += v.y * col_[1].x;
        x += v.z * col_[2].x;
        x += v.w * col_[3].x;

        y  = v.x * col_[0].y;
        y += v.y * col_[1].y;
        y += v.z * col_[2].y;
        y += v.w * col_[3].y;

        z  = v.x * col_[0].z;
        z += v.y * col_[1].z;
        z += v.z * col_[2].z;
        z += v.w * col_[3].z;

        w  = v.x * col_[0].w;
        w += v.y * col_[1].w;
        w += v.z * col_[2].w;
        w += v.w * col_[3].w;

        return vec4(x, y, z, w);
    }

    mat4 operator*(const mat4& m) const {
        return mat4::with_cols(
            (*this) * m.col_[0],
            (*this) * m.col_[1],
            (*this) * m.col_[2],
            (*this) * m.col_[3]
        );
    }

private:
    mat4(const vec4& col0, const vec4& col1,
         const vec4& col2, const vec4& col3)
      : col_{col0, col1, col2, col3} { }
};

}      // namespace px
#endif // _PX_LINALG_HH_
