#ifndef _PX_LINALG_SIMD_HH_
#define _PX_LINALG_SIMD_HH_

// requires: px/simd.hh, px/linalg.hh

namespace px {

struct vec3_4;
struct vec2_4;
inline vec3_4 operator*(const float_4& a, const vec3_4& v);
inline vec2_4 operator*(const float_4& a, const vec2_4& v);

// 4-at-a-time version of vec3
struct vec3_4 {
    float_4 x;
    float_4 y;
    float_4 z;

    // construction
    vec3_4() { }
    vec3_4(const float_4& xv, const float_4& yv, const float_4& zv)
        : x(xv), y(yv), z(zv) { }
    vec3_4(const vec3_4& v) : x(v.x), y(v.y), z(v.z) { }
    vec3_4(const vec3& a, const vec3& b, const vec3& c, const vec3& d)
        : x(float_4(a.x, b.x, c.x, d.x))
        , y(float_4(a.y, b.y, c.y, d.y))
        , z(float_4(a.z, b.z, c.z, d.z)) { }

    explicit vec3_4(const vec3& v)
        : x(v.x, v.x, v.x, v.x)
        , y(v.y, v.y, v.y, v.y)
        , z(v.z, v.z, v.z, v.z) { }

    static vec3_4 zero()    { return vec3_4(vec3::zero()); }
    static vec3_4 x_basis() { return vec3_4(vec3::x_basis()); }
    static vec3_4 y_basis() { return vec3_4(vec3::y_basis()); }
    static vec3_4 z_basis() { return vec3_4(vec3::z_basis()); }

    // assignment
    vec3_4& operator=(const vec3_4& v) { x=v.x; y=v.y; z=v.z; return *this; }

    // access
    vec3 get(int i) const { return vec3(x.at(i), y.at(i), z.at(i)); }
    void set(int i, const vec3& v) { x.at(i)=v.x; y.at(i)=v.y; z.at(i)=v.z; }

    // component-wise arithmetic
    vec3_4 operator-() const { return vec3_4(-x, -y, -z); }
    vec3_4 operator+(const vec3_4& v) const
        { return vec3_4(x+v.x, y+v.y, z+v.z); }
    vec3_4 operator-(const vec3_4& v) const
        { return vec3_4(x-v.x, y-v.y, z-v.z); }
    vec3_4 operator*(const vec3_4& v) const
        { return vec3_4(x*v.x, y*v.y, z*v.z); }
    vec3_4 operator/(const vec3_4& v) const
        { return vec3_4(x/v.x, y/v.y, z/v.z); }

    // destructive component-wise arithmetic
    vec3_4& operator+=(const vec3_4& v) {x+=v.x; y+=v.y; z+=v.z; return *this;}
    vec3_4& operator-=(const vec3_4& v) {x-=v.x; y-=v.y; z-=v.z; return *this;}
    vec3_4& operator*=(const vec3_4& v) {x*=v.x; y*=v.y; z*=v.z; return *this;}
    vec3_4& operator/=(const vec3_4& v) {x/=v.x; y/=v.y; z/=v.z; return *this;}

    // scalar arithmetic
    vec3_4  operator* (const float_4& a) const { return vec3_4(a*x, a*y, a*z); }
    vec3_4  operator/ (const float_4& a) const { return vec3_4(a/x, a/y, a/z); }
    vec3_4& operator*=(const float_4& a) { x*=a; y*=a; z*=a; return *this; }
    vec3_4& operator/=(const float_4& a) { x/=a; y/=a; z/=a; return *this; }

    // dot product
    float_4 dot(const vec3_4& v) const { return x*v.x + y*v.y + z*v.z; }

    // norm
    float_4 norm_sqr() const { return dot(*this); }
    float_4 norm()     const { return norm_sqr().sqrt(); }
    float_4 rnorm()    const { return norm_sqr().rsqrt(); }

    // normalize
    vec3_4  normalize() const { return rnorm() * (*this); }
    vec3_4& normalize_d()     { return *this *= rnorm(); }
};

// 4-at-a-time version of vec2
struct vec2_4 {
    float_4 x, y;

    // construction
    vec2_4() { } // avoid
    vec2_4(const vec2_4& v) : x(v.x), y(v.y) { }
    vec2_4(const float_4& x, const float_4& y) : x(x), y(y) { }
    vec2_4(const vec2& a, const vec2& b, const vec2& c, const vec2& d)
        : x(a.x, b.x, c.x, d.x), y(a.y, b.y, c.y, d.y) { }

    explicit vec2_4(const vec2& a) : x(a.x), y(a.y) { }

    static vec2_4 zero()    { return vec2_4(vec2::zero()); }
    static vec2_4 x_basis() { return vec2_4(vec2::x_basis()); }
    static vec2_4 y_basis() { return vec2_4(vec2::y_basis()); }

    // assignment
    vec2_4& operator=(const vec2_4& v) {
        x = v.x;
        y = v.y;
        return *this;
    }

    // access
    vec2 get(int i) const { return vec2(x.at(i), y.at(i)); }
    void set(int i, const vec2& v) { x.at(i)=v.x; y.at(i)=v.y; }

    // component-wise arithmetic
    vec2_4 operator-() const { return vec2_4(-x, -y); }
    vec2_4 operator+(const vec2_4& v) const { return vec2_4(x+v.x, y+v.y); }
    vec2_4 operator-(const vec2_4& v) const { return vec2_4(x-v.x, y-v.y); }
    vec2_4 operator*(const vec2_4& v) const { return vec2_4(x*v.x, y*v.y); }
    vec2_4 operator/(const vec2_4& v) const { return vec2_4(x/v.x, y/v.y); }

    // destructive component-wise arithmetic
    vec2_4& operator+=(const vec2_4& v) { x+=v.x; y+=v.y; return *this; }
    vec2_4& operator-=(const vec2_4& v) { x-=v.x; y-=v.y; return *this; }
    vec2_4& operator*=(const vec2_4& v) { x*=v.x; y*=v.y; return *this; }
    vec2_4& operator/=(const vec2_4& v) { x/=v.x; y/=v.y; return *this; }

    // scalar arithmetic
    vec2_4  operator*(float_4 a) const { return vec2_4(a*x, a*y); }
    vec2_4  operator/(float_4 a) const { return vec2_4(x/a, y/a); }
    vec2_4& operator*=(float_4 a) { x*=a; y*=a; return *this; }
    vec2_4& operator/=(float_4 a) { x/=a; y/=a; return *this; }

    // dot product
    float_4 dot(const vec2_4& v) const { return x*v.x + y*v.y; }

    // norm
    float_4 norm_sqr() const { return dot(*this); }
    float_4 norm()     const { return norm_sqr().sqrt(); }
    float_4 rnorm()    const { return norm_sqr().rsqrt(); }

    // normalize
    vec2_4  normalize() const { return rnorm() * (*this); }
    vec2_4& normalize_d() { return *this *= rnorm(); }
};

inline vec2_4 operator*(const float_4& a, const vec2_4& v) { return v * a; }
inline vec3_4 operator*(const float_4& a, const vec3_4& v) { return v * a; }

// generic access to vec3_4
template<>
struct simd<vec3> {
    typedef vec3_4 type;
    static type make(const vec3& a, const vec3& b, const vec3& c, const vec3& d) {
        return vec3_4(a, b, c, d);
    }
    static vec3 get(const type& q, int i) { return q.get(i); }
    static void set(type& q, int i, const vec3& v) { q.set(i, v); }
    static type choose(
        const mask_4& m, const type& then, const type& els) {
        typedef simd<float> sf;
        return vec3_4(
            sf::choose(m, then.x, els.x),
            sf::choose(m, then.y, els.y),
            sf::choose(m, then.z, els.z));
    }
};

// generic access to vec2_4
template<>
struct simd<vec2> {
    typedef vec2_4 type;
    static type make(const vec2& a, const vec2& b, const vec2& c, const vec2& d) {
        return vec2_4(a, b, c, d);
    }
    static vec2 get(const type& q, int i) { return q.get(i); }
    static void set(type& q, int i, const vec2& v) { q.set(i, v); }
    static type choose(
        const mask_4& m, const type& then, const type& els) {
        typedef simd<float> sf;
        return vec2_4(
            sf::choose(m, then.x, els.x),
            sf::choose(m, then.y, els.y));
    }
};

}      // namespace px
#endif // _PX_LINALG_SIMD_HH_
