#ifndef _PX_QLINEAR_SIMD_HH_
#define _PX_QLINEAR_SIMD_HH_

// requires: px/simd.hh, px/qlinear.hh

namespace px {

// 4-at-a-time qlinear
template<typename U>
struct qlinear_4 {
    float_4                rw;
    typename simd<U>::type val;

    qlinear_4() { }
    qlinear_4(const qlinear_4<U>& q) : rw(q.rw), val(q.val) { }
    qlinear_4(const float_4& rw, const typename simd<U>::type& u)
        : rw(rw), val(u) { }
    qlinear_4(const qlinear<U>& a, const qlinear<U>& b,
              const qlinear<U>& c, const qlinear<U>& d)
        : rw (a.rw,  b.rw,  c.rw,  d.rw) {
        val = simd<U>::make(a.val, b.val, c.val, d.val);
    }

    explicit qlinear_4(const qlinear<U>& a) : rw(a.rw), val(a.val) { }

    void set(int i, const qlinear<U>& a) {
        rw.set(i, a.rw);
        simd<U>::set(val, i, a.val);
    }

    typename simd<U>::type extract() const { return rw.recip() * val; }
};

template<>
struct qlinear_4<empty_rec_t> {
    qlinear_4() { }
    qlinear_4(const qlinear_4<empty_rec_t>& q) { }
    qlinear_4(const float_4& rw, const empty_rec_t& u) { }
    qlinear_4(const qlinear<empty_rec_t>& a, const qlinear<empty_rec_t>& b,
              const qlinear<empty_rec_t>& c, const qlinear<empty_rec_t>& d) { }
    void set(int i, const qlinear<empty_rec_t>& q) { }
    empty_rec_t extract() const { return empty_rec; }
};

template<typename U>
inline qlinear_4<U> operator-(const qlinear_4<U>& a) {
    return qlinear_4<U>(-a.rw, -a.val);
}

template<typename U>
inline qlinear_4<U> operator+(const qlinear_4<U>& a, const qlinear_4<U>& b) {
    return qlinear_4<U>(a.rw + b.rw, a.val + b.val);
}

template<typename U>
inline qlinear_4<U> operator-(const qlinear_4<U>& a, const qlinear_4<U>& b) {
    return qlinear_4<U>(a.rw - b.rw, a.val - b.val);
}

template<typename U>
inline qlinear_4<U> operator*(const float_4& a, const qlinear_4<U>& b) {
    return qlinear_4<U>(a * b.rw, a * b.val);
}

template<>
inline qlinear_4<empty_rec_t> operator+(const qlinear_4<empty_rec_t>& a,
                                        const qlinear_4<empty_rec_t>& b) {
    return qlinear_4<empty_rec_t>();
}

template<>
inline qlinear_4<empty_rec_t> operator-(const qlinear_4<empty_rec_t>& a,
                                        const qlinear_4<empty_rec_t>& b) {
    return qlinear_4<empty_rec_t>();
}

template<>
inline qlinear_4<empty_rec_t> operator*(
    const float_4& a, const qlinear_4<empty_rec_t>& b) {
    return qlinear_4<empty_rec_t>();
}

template<typename U>
struct simd<qlinear<U> > {
    typedef qlinear_4<U> type;

    static type make(const qlinear<U>& a, const qlinear<U>& b,
                     const qlinear<U>& c, const qlinear<U>& d) {
        return qlinear_4<U>(a, b, c, d);
    }

    static qlinear<U> get(const type& q, int i) {
        return qlinear<U>(q.rw.get(i), simd<U>::get(q.val, i));
    }

    static void set(type& q, int i, const qlinear<U> qlin) { q.set(i, qlin); }

    static type choose(
        const mask_4& m, const type& then, const type& els) {
        typedef simd<float> sf;
        typedef simd<U>     su;
        return qlinear<U>(
            sf::choose(m, then.head_, els.head_),
            su::choose(m, then.tail_, els.tail_));
    }
};

}      // namespace px
#endif // _PX_QLINEAR_SIMD_HH_
