#ifndef _PX_RASTERIZER_HH_
#define _PX_RASTERIZER_HH_

// requires: px/*.hh

#include <cassert>
#include <cstdint>
#include <memory>
#include <limits>

namespace px {

// Macro for creating new shader objects.
// Required make_shader_obj_(...) function is at end of file.
// Use 'varying_t' for the pixel shader varying input type.
// Also! be careful that all by-reference lambda captures live
// for as long as the resulting shader_obj.
#define PX_MACRO_APPEND(a, b)  a##b
#define PX_MACRO_APPEND2(a, b) PX_MACRO_APPEND(a, b)
#define PX_NAME_LINENO(a)      PX_MACRO_APPEND2(a, __LINE__)
#define PX_MAKE_SHADER_OBJ(ctx, vs, ps) ({                                \
    auto PX_NAME_LINENO(PXMSO_vertex_shader_) = (vs);                     \
    typedef typename simd<                                                \
        typename field_split<                                             \
            L_pos_t,                                                      \
            decltype(PX_NAME_LINENO(PXMSO_vertex_shader_)(                \
                Uni(), vec3(), VAttr()))                                  \
        >::nonfocus_t                                                     \
    >::type varying_t;                                                    \
    make_shader_obj_((ctx), PX_NAME_LINENO(PXMSO_vertex_shader_), (ps)); })

// Special fields for vertex shader output
PX_REC_DEF_FIELD(pos);      // homogeneous coordinates (main output)
PX_REC_DEF_FIELD(qlinear);  // request for rational linear interpolation

enum cull_face_t { CULL_FRONT, CULL_BACK };

template<typename Uni, typename VAttr, typename Px>
class context;

template<typename Uni, typename VAttr, typename Px>
class shader_obj {
protected:
    context<Uni, VAttr, Px>& ctx_;

public:
    typedef typename rec_arrays<VAttr>::type attr_arrays_t;

    shader_obj(context<Uni, VAttr, Px>& ctx) : ctx_(ctx) { }
    virtual ~shader_obj() { }

    context<Uni, VAttr, Px>& get_context() { return ctx_; }

    virtual void draw_tris(
        int n, const vec3*     verts,    // verts[0..n-1]
        int f, const uint16_t* tris,     // tris[0..3f-1]
        const attr_arrays_t&   attrs)=0; // attrs[var][0..n-1]

    // draw splits from last call to draw_tris() onto color buffer
    virtual void draw_splits(const Px& color)=0;
};

class shader_obj_id {
    unsigned i_;
public:
    shader_obj_id(unsigned i) : i_(i) { }
    shader_obj_id(const shader_obj_id& so_id) : i_(so_id.i_) { }
    unsigned index() const { return i_; }
};

template<typename Uni, typename VAttr, typename Px>
class context {
    // state
    worker_pool& wpool_;
    Uni   uni_;
    bool  enable_cull_;
    bool  front_ccw_;
    bool  wireframe_;
    bool  color_write_;
    bool  depth_test_;
    bool  depth_write_;
    float near_;
    int   clip_x_, clip_y_, clip_w_, clip_h_;

    // shaders
    std::vector<std::unique_ptr<shader_obj<Uni, VAttr, Px> > >
        all_shaders_;
    shader_obj_id cur_shader_;

    // buffers
    buffer<Px>*    tgt_;
    buffer<float>* zbuf_;
    int            w_, h_;

public:
    typedef typename shader_obj<Uni, VAttr, Px>::attr_arrays_t attr_arrays_t;

    context(worker_pool& wpool, const Uni& uni)
        : wpool_(wpool)
        , uni_(uni)
        , enable_cull_(true)
        , front_ccw_(true)
        , wireframe_(false)
        , color_write_(true)
        , depth_test_(true)
        , depth_write_(true)
        , near_(1.f)
        , clip_x_(0)
        , clip_y_(0)
        , clip_w_(0)
        , clip_h_(0)
        , all_shaders_()
        , cur_shader_(-1)
        , tgt_(NULL)
        , zbuf_(NULL)
        , w_(-1)
        , h_(-1) {
        assert(wpool_.wasinit());
    }

    worker_pool& get_worker_pool() { return wpool_; }

    const Uni& uniform() const { return uni_; }
    Uni& uniform() { return uni_; }

    // NB. transfers ownership to context
    shader_obj_id register_shader(
        std::unique_ptr<shader_obj<Uni, VAttr, Px> > so_ptr) {
        assert(so_ptr != NULL);
        assert(&so_ptr->get_context() == this);
        all_shaders_.push_back(std::move(so_ptr));
        return shader_obj_id(all_shaders_.size() - 1);
    }

    shader_obj_id current_shader() const { return cur_shader_; }

    void bind_shader(shader_obj_id so_id) {
        assert(so_id.index() >= 0 && so_id.index() < all_shaders_.size());
        cur_shader_ = so_id;
    }

    void bind_buffers(buffer<Px>* color, buffer<float>* depth) {
        assert(color != NULL);
        assert(depth != NULL);
        w_ = color->width();
        h_ = color->height();
        assert(depth->width()  == w_);
        assert(depth->height() == h_);
        tgt_  = color;
        zbuf_ = depth;
        set_clip_region(0, 0, w_, h_);
    }

    bool cull_enabled() const { return enable_cull_; }
    void enable_cull(bool cull) { enable_cull_ = cull; }

    cull_face_t cull_face() const { return front_ccw_ ? CULL_BACK : CULL_FRONT; }
    void set_cull_face(cull_face_t face) { front_ccw_ = face == CULL_BACK; }

    float near_distance() const { return near_; }
    void set_near_distance(float near) { near_ = near; }

    void clip_region(int& x, int& y, int& w, int& h) {
        x = clip_x_;
        y = clip_y_;
        w = clip_w_;
        h = clip_h_;
    }

    void set_clip_region(int x, int y, int w, int h) {
        clip_x_ = x;
        clip_y_ = y;
        clip_w_ = w;
        clip_h_ = h;
    }

    bool wireframe() const { return wireframe_; }
    void set_wireframe(bool enable) { wireframe_ = enable; }

    bool depth_test_enabled() const { return depth_test_; }
    void enable_depth_test(bool test) { depth_test_ = test; }

    bool depth_write_enabled() const { return depth_write_; }
    void enable_depth_write(bool write) { depth_write_ = write; }

    bool color_write_enabled() const { return color_write_; }
    void enable_color_write(bool write) { color_write_ = write; }

    int width() const { return w_; }
    int height() const { return h_; }

    buffer<Px>*    colorbuf() const { return tgt_; }
    buffer<float>* depthbuf() const { return zbuf_; }

    void draw_tris(
        int n, const vec3*     verts,    // verts[0..n-1]
        int f, const uint16_t* tris,     // tris[0..3f-1]
        const attr_arrays_t&   attrs) {  // attrs[var][0..n-1]
        assert(cur_shader_.index() != (unsigned)-1);
        all_shaders_[cur_shader_.index()]->draw_tris(n, verts, f, tris, attrs);
    }

    // Draw splits from last call to draw_tris() onto color buffer.
    // The splits are associated with THE SHADER, so they will
    // change if shaders are rebound to a different object.
    void draw_splits(const Px& color) {
        assert(cur_shader_.index() != (unsigned)-1);
        all_shaders_[cur_shader_.index()]->draw_splits(color);
    }
};

template<typename Uni, typename VAttr, typename VS, typename PS, typename Px>
class shader_obj_impl : public shader_obj<Uni, VAttr, Px> {
    // early member variables (for decltype)
    VS  vs_;
    PS  ps_;

public:
    // typedefs
    typedef typename shader_obj<Uni, VAttr, Px>::attr_arrays_t attr_arrays_t;
    typedef decltype(vs_(Uni(), vec3(), VAttr())) vsout_t;
    typedef field_split<L_qlinear_t, vsout_t>     qlin_partn_t;
    typedef typename qlin_partn_t::focus_t        qlin_bare_t;
    typedef qlinear<qlin_bare_t>                  qlin_t;
    typedef typename qlin_partn_t::nonfocus_t     lin_t;
    typedef field_split<L_pos_t, lin_t>           pos_partn_t;
    typedef typename pos_partn_t::nonfocus_t      vary_t;

private:
    // local types --
    struct vertinfo {
        vec4   pos;
        qlin_t qlin;
        vary_t vary;
        vertinfo() { }
        vertinfo(const vec4& p, const qlin_t& q, const vary_t& v)
            : pos(p), qlin(q), vary(v) { }
    };

    struct rect {
        int xlo, ylo;
        int xhi, yhi;
        int width()  const { return xhi - xlo; }
        int height() const { return yhi - ylo; }
        int area() const { return width() * height(); }
    };

    // The type of clipped triangles. The aabb is for splitting.
    // We clip in homogeneous space.
    // Each clip introduces at most one extra vertex,
    // so starting with a triangle and clipping 6 times,
    // we always get at most 9 vertices.
    struct clipped_tri { int k; vertinfo verts[9]; rect aabb; };

    // Lines and wireframe triangles
    struct anyline { vertinfo ends[2]; };
    struct wireframe_tri { int k; anyline lines[3]; };

    // trapezoid types
    struct corner {
        float  x;
        float  z;
        qlin_t qlin;
        vary_t vary;
        corner() { }
        corner(float x, float z, const qlin_t& q, const vary_t& v)
            : x(x), z(z), qlin(q), vary(v) { }
        corner(const vertinfo& vi)
            : x(vi.pos.x), z(vi.pos.z), qlin(vi.qlin), vary(vi.vary) { }
        corner operator+(const corner& c) const {
            return corner(x + c.x, z + c.z, qlin + c.qlin, vary + c.vary);
        }
        corner operator-(const corner& c) const {
            return corner(x - c.x, z - c.z, qlin - c.qlin, vary - c.vary);
        }
        corner operator*(float a) const {
            return corner(a * x, a * z, a * qlin, a * vary);
        }
    };

    struct scanline {
        int    y;
        corner left;
        corner right;
        scanline() { }
        scanline(int y, const corner& left, const corner& right)
            : y(y), left(left), right(right) { }
    };

    template<typename Derived>
    struct clipper_base {
        // specialized clipping code wrappers (for short calls)
        float coord_(const vec4& v) const {
            return static_cast<const Derived*>(this)->coord(v);
        }
        float bound_(const vec4& v) const {
            return static_cast<const Derived*>(this)->bound(v);
        }
        bool  inside_(float x, float bound) const {
            return static_cast<const Derived*>(this)->inside(x, bound);
        }
        void  rescale_w_(vertinfo& a, const vertinfo& b) const {
            return static_cast<const Derived*>(this)->rescale_w(a, b);
        }
        void  lerp4_fix_(vec4& v) const {
            return static_cast<const Derived*>(this)->lerp4_fix(v);
        }

        // shared clipping code
        bool point_inside(const vec4& v) {
            return inside_(coord_(v), bound_(v));
        }

        bool clip_line(anyline& line) {
            bool bad[2];
            bad[0] = !point_inside(line.ends[0].pos);
            bad[1] = !point_inside(line.ends[1].pos);
            if (bad[0] && bad[1])
                return false;

            int bad_end = -1;
            if (bad[0])
                bad_end = 0;
            else if (bad[1])
                bad_end = 1;

            if (bad_end != -1) {
                int good_end = 1 - bad_end;
                line.ends[bad_end] = lerp4_boundary(
                    line.ends[bad_end], line.ends[good_end]);
            }

            return true;
        }

        int clip_tri(const vertinfo *poly, int k, vertinfo out[9]) {
            int kprime = 0;
            const vertinfo* last = &poly[k];
            const vertinfo* prev = &poly[0];
            bool  prev_in = point_inside(prev->pos);
            bool  cur_in; // set in clip_edge
            for (const vertinfo* cur = &poly[1]; cur != last; cur++) {
                clip_edge(*prev, *cur, prev_in, cur_in, out, kprime);
                prev    = cur;
                prev_in = cur_in;
            }
            clip_edge(*prev, poly[0], prev_in, cur_in, out, kprime);
            return kprime;
        }

        void clip_edge(const vertinfo& prev, const vertinfo& cur,
                       bool& prev_in, bool& cur_in,
                       vertinfo out[9], int& kprime) {
            cur_in = point_inside(cur.pos);
            if (cur_in) {
                if (!prev_in)
                    out[kprime++] = lerp4_boundary(prev, cur);
                out[kprime++] = cur;
            } else if (prev_in) {
                out[kprime++] = lerp4_boundary(prev, cur);
            }
        }

        vertinfo lerp4_boundary(vertinfo a, const vertinfo& b) {
            rescale_w_(a, b);
            float ax = coord_(a.pos), bx = coord_(b.pos);
            float t  = (bound_(b.pos) - ax) / (bx - ax);
            float s  = 1.f - t;
            a.pos.x *= s;
            a.pos.y *= s;
            a.pos.z *= s;
            a.qlin   = s * a.qlin;
            a.vary   = s * a.vary;
            a.pos.x += t * b.pos.x;
            a.pos.y += t * b.pos.y;
            a.pos.z += t * b.pos.z;
            a.qlin   = a.qlin + t * b.qlin;
            a.vary   = a.vary + t * b.vary;
            lerp4_fix_(a.pos);
            return a;
        }
    };

    struct near_clipper : public clipper_base<near_clipper> {
        float near;
        float coord(const vec4& v) const { return v.z; }
        float bound(const vec4& v) const { return -near; }
        bool  inside(float x, float bound) const { return x >= bound; }
        void  rescale_w(vertinfo& a, const vertinfo& b) const { }
        void  lerp4_fix(vec4& v) const { v.z = -near; v.w = near; }
    };

    template<typename Derived>
    struct clipper : public clipper_base<Derived> {
        int i; // which component
        float factor;

        clipper() { }
        clipper(float factor) : factor(factor) { }

        float coord(const vec4& v) const { return v.coords[i]; }
        float bound(const vec4& v) const { return factor * v.w; }
        void  lerp4_fix(vec4& v) const { }
        void  rescale_w(vertinfo& a, const vertinfo& b) const {
            float ratio = b.pos.w / a.pos.w;
            a.pos *= ratio;
            a.qlin = ratio * a.qlin;
        }
    };

    struct low_clipper : public clipper<low_clipper> {
        low_clipper() : clipper<low_clipper>(-1.f) { }
        bool inside(float x, float bound) const { return x >= bound; }
    };

    struct high_clipper : public clipper<high_clipper> {
        high_clipper() : clipper<high_clipper>(1.f) { }
        bool inside(float x, float bound) const { return x <= bound; }
    };

    // member variables --

    // interemediate buffers for draw_tris()
    int      vsout_cap_;
    vsout_t* vsout_;

    // clipped triangle buffers
    int           clipped_cap_;
    clipped_tri*  clipped_;
    clipped_tri** clip_bufs_;

    // wireframe triangle buffers
    int             wfcap_;
    wireframe_tri*  wftri_;
    wireframe_tri** wftri_bufs_;

    // splits retained from draw_tris() for debugging/visualization purposes()
    rect* splits_;
    int   nsplits_;

public:
    shader_obj_impl(context<Uni, VAttr, Px>& ctx,
                    const VS& vshader, const PS& pxshader)
        : shader_obj<Uni, VAttr, Px>(ctx)
        , vs_(vshader)
        , ps_(pxshader)
        , vsout_cap_(256)
        , vsout_(new vsout_t[vsout_cap_])
        , clipped_cap_(256)
        , clipped_(new clipped_tri[clipped_cap_])
        , clip_bufs_(new clipped_tri*[ctx.get_worker_pool().count()])
        , wfcap_(256)
        , wftri_(new wireframe_tri[wfcap_])
        , wftri_bufs_(new wireframe_tri*[ctx.get_worker_pool().count()])
        , splits_(new rect[ctx.get_worker_pool().count() - 1])
        , nsplits_(0) {
        int nthreads = ctx.get_worker_pool().count();
        int clip_max_bufsiz = clipped_cap_ / nthreads + nthreads - 1;
        int wf_max_bufsiz = wfcap_ / nthreads + nthreads - 1;
        for (int i = 0; i < nthreads; i++) {
            clip_bufs_[i]  = new clipped_tri[clip_max_bufsiz];
            wftri_bufs_[i] = new wireframe_tri[wf_max_bufsiz];
        }
    }

    ~shader_obj_impl() {
        delete[] vsout_;
        delete[] clipped_;
        delete[] splits_;
        int nthreads = this->ctx_.get_worker_pool().count();
        for (int i = 0; i < nthreads; i++) {
            delete[] clip_bufs_[i];
            delete[] wftri_bufs_[i];
        }
        delete[] clip_bufs_;
        delete[] wftri_bufs_;
    }

    virtual void draw_tris(
        int n, const vec3*     verts,   // verts[0..n-1]
        int f, const uint16_t* tris,    // tris[0..3f-1]
        const attr_arrays_t&   attrs);  // attrs[var][0..n-1]

    // draw splits from last call to draw_tris() onto color buffer
    virtual void draw_splits(const Px& color);

private:
    void transform_verts(int n, const vec3* verts, const attr_arrays_t& attrs);
    void transform_some_verts(int a, int b, const vec3* verts,
                              const attr_arrays_t& attrs);
    int  transform_tris(int f, const uint16_t* tris);
    int  transform_some_tris(int a, int b, const uint16_t* tris, clipped_tri* buf);
    void draw_clipped(int f, const rect& bound);

    void draw_tris_wireframe(
        int f, const uint16_t* tris,   // tris[0..3f-1]
        const attr_arrays_t&   attrs); // attrs[var][0..n-1]
    int  transform_tris_wireframe(int f, const uint16_t* tris);
    int  transform_some_tris_wireframe(int a, int b, const uint16_t* tris, wireframe_tri* buf);
    void draw_screen_line(const anyline& line);

    void screen_split_init(
        int f, const uint16_t* tris, rect* bounds, float* weights);

    bool screen_split1(
        int f, const uint16_t* tris, rect* bounds, float* weights, int i);

    template<typename Lo, typename Hi>
    bool screen_split_subrect(
        const Lo& lo, const Hi& hi,
        int f, const uint16_t* tris, rect* bounds, float* weights,
        int i, int k) {
        rect& to_split = bounds[k];
        assert(to_split.xlo <= to_split.xhi);
        assert(to_split.ylo <= to_split.yhi);

        float split_x    = 0.f;
        float total_area = 0.f;
        for (int j = 0; j < f; j++) {
            rect aabb = clipped_[j].aabb;
            aabb.xlo = clamp(to_split.xlo, to_split.xhi, aabb.xlo);
            aabb.xhi = clamp(to_split.xlo, to_split.xhi, aabb.xhi);
            aabb.ylo = clamp(to_split.ylo, to_split.yhi, aabb.ylo);
            aabb.yhi = clamp(to_split.ylo, to_split.yhi, aabb.yhi);
            assert(aabb.xlo <= aabb.xhi);
            assert(aabb.ylo <= aabb.yhi);

            float cur_area = aabb.area();
            split_x    += cur_area * (lo(aabb) + .5f * (hi(aabb) - lo(aabb)));
            total_area += cur_area;
        }

        split_x /= 4.f * total_area;
        int x = 4 * int(split_x + .5f);
        if (x <= lo(to_split) || x >= hi(to_split))
            return false;

        rect new_rect = to_split;
        hi(to_split) = x;
        lo(new_rect) = x;
        bounds[i]      = new_rect;
        weights[k] *= .5f;
        weights[i] = weights[k];
        splits_[i-1] = to_split;
        lo(splits_[i-1]) = hi(splits_[i-1]) = x;
        return true;
    }

    bool screen_split_horiz(
        int f, const uint16_t* tris, rect* bounds, float* weights,
        int i, int k) {
        return screen_split_subrect(
            [](rect& r) -> int& { return r.xlo; },
            [](rect& r) -> int& { return r.xhi; },
            f, tris, bounds, weights, i, k);
    }

    bool screen_split_verti(
        int f, const uint16_t* tris, rect* bounds, float* weights,
        int i, int k) {
        return screen_split_subrect(
            [](rect& r) -> int& { return r.ylo; },
            [](rect& r) -> int& { return r.yhi; },
            f, tris, bounds, weights, i, k);
    }

    vertinfo split_vsout(const vsout_t& vsout) {
        qlin_t qlin = mkqlinear(1.f, qlin_partn_t::focus(vsout));
        lin_t  lin  = qlin_partn_t::nonfocus(vsout);
        vec4   pos  = pos_partn_t::focus(lin);
        vary_t vary = pos_partn_t::nonfocus(lin);
        return vertinfo(pos, qlin, vary);
    }

    int clip(vertinfo tri[3], vertinfo out[9]) {
        vertinfo tmp[9];
        near_clipper c_near;
        low_clipper  c_low;
        high_clipper c_high;

        // clip near
        c_near.near = this->ctx_.near_distance();
        int k = c_near.clip_tri(tri, 3, tmp);
        if (k == 0)
            return 0;

        // clip far
        c_high.i = 2;
        k = c_high.clip_tri(tmp, k, out);
        if (k == 0)
            return 0;

        // x, y clipping is subject to clipping region
        int low_x, low_y, clip_w, clip_h;
        this->ctx_.clip_region(low_x, low_y, clip_w, clip_h);
        int high_x = low_x + clip_w;
        int high_y = low_y + clip_h;

        // note y is inverted
        float a_x = -1.f + low_x * 2.f / this->ctx_.width();
        float a_y = 1.f - high_y * 2.f / this->ctx_.height();
        float b_x = -1.f + high_x * 2.f / this->ctx_.width();
        float b_y = 1.f - low_y * 2.f / this->ctx_.height();

        float low_bound[2]  = { a_x, a_y };
        float high_bound[2] = { b_x, b_y };

        // clip x, y
        for (int i = 0; i < 2; i++) {
            c_low.i = i;
            c_low.factor = low_bound[i];
            k = c_low.clip_tri(out, k, tmp);
            if (k == 0)
                return 0;

            c_high.i = i;
            c_high.factor = high_bound[i];
            k = c_high.clip_tri(tmp, k, out);
            if (k == 0)
                return 0;
        }

        assert(k >= 3 && k <= 9);
        return k;
    }

    bool reorder_cull(vertinfo* tri, int k) {
        // find min vertex by y-coordnate,
        // which is the top because y is inverted
        int top = 0;
        int i;
        for (i = 1; i < k; i++) {
            if (tri[i].pos.y < tri[top].pos.y)
                top = i;
        }

        // rotate min vertex into first position
        vertinfo tmp[k];
        for (i = 0; i < top; i++)
            tmp[i] = tri[i];

        int j;
        for (j = 0; i < k; i++, j++)
            tri[j] = tri[i];

        for (i = 0; i < top; i++, j++)
            tri[j] = tmp[i];

        // put the 'left' edge at the beginning, 'right' edge at the end
        float ny = tri[0].pos.x - tri[1].pos.x;
        float nx = tri[1].pos.y - tri[0].pos.y;
        float c  = -nx * tri[0].pos.x - ny * tri[0].pos.y;

        // check winding
        bool ccw = nx * tri[k-1].pos.x + ny * tri[k-1].pos.y + c >= 0;
        if (this->ctx_.cull_enabled() && ccw != (this->ctx_.cull_face() == CULL_BACK))
            return false;

        if (!ccw) {
            // drawing code assumes ccw vertices; need to reverse
            for (i = 1, j = k-1; i < j; i++, j--)
                std::swap(tri[i], tri[j]);
        }
        return true;
    }

    // draws from top up to, but not including, btm
    void draw_trapezoid(
        const scanline& top, const scanline& btm, const rect& bound) {
        float  rydiff = 1.f / (btm.y - top.y);
        corner dleft  = (btm.left  - top.left)  * rydiff;
        corner dright = (btm.right - top.right) * rydiff;
        scanline cur(top.y, top.left, top.right);

        // check that y-clipping is good
        assert(top.y >= 0 && top.y < this->ctx_.height());
        assert(btm.y >= 0 && btm.y < this->ctx_.height());

        while (cur.y < btm.y) {
            if (cur.y >= bound.ylo && cur.y < bound.yhi)
                draw_scanline(cur, bound);
            cur.left  = cur.left + dleft;
            cur.right = cur.right + dright;
            cur.y++;
        }
    }

    void draw_scanline(const scanline& line, const rect& bound) {
        int    lx     = int(line.left.x);
        int    rx     = int(line.right.x);
        float  rxdiff = 1.f / (rx - lx);
        float  dz     = rxdiff * (line.right.z    - line.left.z);
        qlin_t dqlin  = rxdiff * (line.right.qlin - line.left.qlin);
        vary_t dvary  = rxdiff * (line.right.vary - line.left.vary);

        // check that x-clipping is good
        assert(lx >= 0 && lx < this->ctx_.width());
        assert(rx >= 0 && rx < this->ctx_.width());

        // toss completely wrong scanlines
        if (rx < bound.xlo || lx >= bound.xhi)
            return;

        float z0 = line.left.z;
        float z1 = z0 + dz;
        float z2 = z1 + dz;
        float z3 = z2 + dz;
        float_4 z(z0, z1, z2, z3);

        qlin_t qlin0 = line.left.qlin;
        qlin_t qlin1 = qlin0 + dqlin;
        qlin_t qlin2 = qlin1 + dqlin;
        qlin_t qlin3 = qlin2 + dqlin;
        typename simd<qlin_t>::type qlin =
            simd<qlin_t>::make(qlin0, qlin1, qlin2, qlin3);

        vary_t vary0 = line.left.vary;
        vary_t vary1 = vary0 + dvary;
        vary_t vary2 = vary1 + dvary;
        vary_t vary3 = vary2 + dvary;
        typename simd<vary_t>::type vary =
            simd<vary_t>::make(vary0, vary1, vary2, vary3);

        // increments
        float_4                     dz4(dz);
        typename simd<qlin_t>::type dqlin4 = simd_make1(dqlin);
        typename simd<vary_t>::type dvary4 = simd_make1(dvary);

        // clip to (left side of) bound
        if (lx < bound.xlo) {
            float_4 diff(float(bound.xlo - lx));
            z   += diff * dz4;
            qlin = qlin + diff * dqlin4;
            vary = vary + diff * dvary4;
            lx = bound.xlo;
        }

        // need x to be a multiple of 4
        int lx_align = lx & 0x3;
        float_4 lx_align_f4((float)lx_align);
        z    -= lx_align_f4 * dz4;
        qlin  = qlin - lx_align_f4 * dqlin4;
        vary  = vary - lx_align_f4 * dvary4;

        // increment by multiples of four
        dz4    *= float_4(4.f);
        dqlin4  = float_4(4.f) * dqlin4;
        dvary4  = float_4(4.f) * dvary4;

        // draw (clipping to right side)
        int rx_lim = std::min(rx, bound.xhi - 1);
        int_4 lx4(lx), rx4(rx_lim);
        for (int x = lx - lx_align; x <= rx_lim; x += 4) {
            int_4  x4(x, x+1, x+2, x+3);
            mask_4 ok = (x4 >= lx4) & (x4 <= rx4);
            draw_pixel4(ok, x, x4, line.y, z, qlin, vary);
            z    += dz4;
            qlin  = qlin + dqlin4;
            vary  = vary + dvary4;
        }
    }

    void draw_pixel4(const mask_4& mask, int x, const int_4& x4, int y,
                     float_4 z,
                     const typename simd<qlin_t>::type& qlin,
                     const typename simd<vary_t>::type& vary) {
        // The conditionals here (color_write_, depth_{test,write}_)
        // seem to be okay. They don't change for an entire draw() call,
        // so I imagine the CPU learns to predict these quickly.
        typedef typename simd<Px>::type pxs_t;
        mask_4 write = mask;
        float_4* z_ref    = &this->ctx_.depthbuf()->at_quad(x, y);
        float_4  z_orig   = *z_ref;
        if (this->ctx_.depth_test_enabled())
            write &= z < z_orig;

        if (!write.any())
            return;

        pxs_t* pxs_ref  = &this->ctx_.colorbuf()->at_quad(x, y);
        pxs_t  pxs_orig = *pxs_ref;
        auto   ps_arg   = (vary, L_qlinear = qlin.extract());

        // run shader - this writes z and pxs
        pxs_t pxs(pxs_orig);
        ps_(this->ctx_.uniform(), ps_arg, x4, int_4(y), z, pxs);

        // write results
        if (this->ctx_.color_write_enabled())
            *pxs_ref = simd<Px>::choose(write, pxs, pxs_orig);
        if (this->ctx_.depth_write_enabled())
            *z_ref = simd<float>::choose(write, z, z_orig);
    }

    void ndc2dc(vec4& p) {
        p.x *= .5f;
        p.y *= -.5f;
        p.x += .5f;
        p.y += .5f;
        p.x *= this->ctx_.width();
        p.y *= this->ctx_.height();
        // just to be extra safe
        p.x  = clamp(0.f, this->ctx_.width()  - .05f, p.x);
        p.y  = clamp(0.f, this->ctx_.height() - .05f, p.y);
    }
};

template<typename Uni, typename VAttr, typename VS, typename PS, typename Px>
void shader_obj_impl<Uni, VAttr, VS, PS, Px>::transform_verts(
    int n, const vec3* verts, const attr_arrays_t& attrs) {
    // ensure enough room for transformed vertices
    if (n > vsout_cap_) {
        while (n > vsout_cap_)
            vsout_cap_ *= 2;

        delete[] vsout_;
        vsout_ = new vsout_t[vsout_cap_];
    }

    // transform vertices
    int nthreads = this->ctx_.get_worker_pool().count();
    int inds[nthreads + 1];
    for (int i = 0; i < nthreads; i++)
        inds[i] = i * n / nthreads;
    inds[nthreads] = n;

    auto that = this;
    this->ctx_.get_worker_pool().distribute_work([&](int i) {
        that->transform_some_verts(inds[i], inds[i+1], verts, attrs);
    });
}

template<typename Uni, typename VAttr, typename VS, typename PS, typename Px>
void shader_obj_impl<Uni, VAttr, VS, PS, Px>::transform_some_verts(
    int a, int b, const vec3* verts, const attr_arrays_t& attrs) {
    for (int i = a; i < b; i++)
        vsout_[i] = vs_(
            this->ctx_.uniform(), verts[i], rec_arrays<VAttr>::index(attrs, i));
}

template<typename Uni, typename VAttr, typename VS, typename PS, typename Px>
int shader_obj_impl<Uni, VAttr, VS, PS, Px>::transform_tris(
    int f, const uint16_t* tris) {
    int nthreads = this->ctx_.get_worker_pool().count();

    // ensure enough room for transformed (clipped) triangles
    if (f > clipped_cap_) {
        while (f > clipped_cap_)
            clipped_cap_ *= 2;

        delete[] clipped_;
        clipped_ = new clipped_tri[clipped_cap_];
        int max_bufsiz = clipped_cap_ / nthreads + nthreads - 1;
        for (int i = 0; i < nthreads; i++) {
            delete[] clip_bufs_[i];
            clip_bufs_[i] = new clipped_tri[max_bufsiz];
        }
    }

    // results go into here (and clip_bufs_)
    int counts[nthreads];
    int parts[nthreads + 1];
    for (int i = 0; i < nthreads; i++)
        parts[i] = i * f / nthreads;
    parts[nthreads] = f;

    // clip
    auto that = this;
    this->ctx_.get_worker_pool().distribute_work([&](int thread) {
        int start = parts[thread], end = parts[thread+1];
        clipped_tri* buf = that->clip_bufs_[thread];
        counts[thread] = that->transform_some_tris(start, end, tris, buf);
    });

    // get a total, and cumulative counts
    int total_count = 0;
    int cum_counts[nthreads]; // cum_counts[t] counst up to and including t
    for (int i = 0; i < nthreads; i++)
        cum_counts[i] = (total_count += counts[i]);

    // copy data into clipped_
    this->ctx_.get_worker_pool().distribute_work([&](int thread) {
        int start = thread == 0 ? 0 : cum_counts[thread - 1];
        int count = counts[thread];
        clipped_tri* buf = that->clip_bufs_[thread];
        for (int i = 0; i < count; i++)
            clipped_[start + i] = buf[i];
    });

    return total_count;
}

template<typename Uni, typename VAttr, typename VS, typename PS, typename Px>
int shader_obj_impl<Uni, VAttr, VS, PS, Px>::transform_some_tris(
    int start, int end, const uint16_t* tris, clipped_tri* buf) {

    int inf       = std::numeric_limits<int>::max();
    int minus_inf = std::numeric_limits<int>::min();

    int count = 0;
    for (int i = start; i < end; i++) {
        vertinfo tri[3];
        int i3 = 3 * i;
        for (int j = 0; j < 3; j++)
            tri[j] = split_vsout(vsout_[tris[i3 + j]]);

        buf[count].k = clip(tri, buf[count].verts);
        if (buf[count].k == 0)
            continue; // entirely outside clipping volume

        // complete the projection (i.e. divide)
        for (int j = 0; j < buf[count].k; j++) {
            vertinfo& vinfo = buf[count].verts[j];
            float rw = 1.f / vinfo.pos.w;
            vinfo.qlin = rw * vinfo.qlin;
            vinfo.pos.x *= rw;
            vinfo.pos.y *= rw;
            vinfo.pos.z *= rw;
            ndc2dc(vinfo.pos); // put x and y in terms of pixels
        }

        // impose a convenient order on the vertices,
        // and cull if appropriate
        if (!reorder_cull(buf[count].verts, buf[count].k))
            continue;

        // compute axis-aligned bounding box
        rect& aabb = buf[count].aabb;
        aabb.xlo = inf;
        aabb.xhi = minus_inf;
        aabb.ylo = inf;
        aabb.yhi = minus_inf;
        for (int j = 0; j < buf[count].k; j++) {
            const vertinfo& vi = buf[count].verts[j];
            int x(vi.pos.x + .5f), y(vi.pos.y + .5f);
            aabb.xlo = std::min(aabb.xlo, x);
            aabb.xhi = std::max(aabb.xhi, x);
            aabb.ylo = std::min(aabb.ylo, y);
            aabb.yhi = std::max(aabb.yhi, y);
        }

        count++;
    }

    return count;
}

template<typename Uni, typename VAttr, typename VS, typename PS, typename Px>
int shader_obj_impl<Uni, VAttr, VS, PS, Px>::transform_tris_wireframe(
    int f, const uint16_t* tris) {
    int nthreads = this->ctx_.get_worker_pool().count();

    // ensure enough room for transformed (clipped) triangles
    if (f > wfcap_) {
        while (f > wfcap_)
            wfcap_ *= 2;

        delete[] wftri_;
        wftri_ = new wireframe_tri[wfcap_];
        int max_bufsiz = wfcap_ / nthreads + nthreads - 1;
        for (int i = 0; i < nthreads; i++) {
            delete[] wftri_bufs_[i];
            wftri_bufs_[i] = new wireframe_tri[max_bufsiz];
        }
    }

    // results go into here (and wftri_bufs_)
    int counts[nthreads];
    int parts[nthreads + 1];
    for (int i = 0; i < nthreads; i++)
        parts[i] = i * f / nthreads;
    parts[nthreads] = f;

    // clip
    auto that = this;
    this->ctx_.get_worker_pool().distribute_work([&](int thread) {
        int start = parts[thread], end = parts[thread+1];
        wireframe_tri* buf = that->wftri_bufs_[thread];
        counts[thread] = that->transform_some_tris_wireframe(
            start, end, tris, buf);
    });

    // get a total, and cumulative counts
    int total_count = 0;
    int cum_counts[nthreads]; // cum_counts[t] counst up to and including t
    for (int i = 0; i < nthreads; i++)
        cum_counts[i] = (total_count += counts[i]);

    // copy data into clipped_
    this->ctx_.get_worker_pool().distribute_work([&](int thread) {
        int start = thread == 0 ? 0 : cum_counts[thread - 1];
        int count = counts[thread];
        wireframe_tri* buf = that->wftri_bufs_[thread];
        for (int i = 0; i < count; i++)
            wftri_[start + i] = buf[i];
    });

    return total_count;
}

template<typename Uni, typename VAttr, typename VS, typename PS, typename Px>
int shader_obj_impl<Uni, VAttr, VS, PS, Px>::transform_some_tris_wireframe(
    int start, int end, const uint16_t* tris, wireframe_tri* buf) {

    int count = 0;
    for (int i = start; i < end; i++) {
        vertinfo tri[3];
        int i3 = 3 * i;
        for (int j = 0; j < 3; j++)
            tri[j] = split_vsout(vsout_[tris[i3 + j]]);

        buf[count].k = 0;
        for (int j = 0; j < 3; j++) {
            anyline line = { { tri[j], tri[j == 2 ? 0 : j+1] } };

            near_clipper c_near;
            low_clipper  c_low;
            high_clipper c_high;

            // clip against near plane
            c_near.near = this->ctx_.near_distance();
            if (!c_near.clip_line(line))
                goto next_line;

            // far plane
            c_high.i = 3;
            if (!c_high.clip_line(line))
                goto next_line;

            // left/right/top/bottom
            for (int c = 0; c < 2; c++) {
                c_low.i = c_high.i = c;
                if (!c_low.clip_line(line))
                    goto next_line;
                if (!c_high.clip_line(line))
                    goto next_line;
            }

            // line survived clipping, add it!
            buf[count].lines[buf[count].k++] = line;

        next_line: ;
        }

        if (buf[count].k < 2)
            continue; // entirely outside clipping volume

        // complete the projection (i.e. divide)
        for (int j = 0; j < buf[count].k; j++) {
            for (int k = 0; k < 2; k++) {
                vertinfo& vinfo = buf[count].lines[j].ends[k];
                float rw = 1.f / vinfo.pos.w;
                vinfo.qlin = rw * vinfo.qlin;
                vinfo.pos.x *= rw;
                vinfo.pos.y *= rw;
                vinfo.pos.z *= rw;
                ndc2dc(vinfo.pos); // put x and y in terms of pixels
            }
        }

        // re-implement back-face culling -
        // can't easily share the code here... :(
        if (this->ctx_.cull_enabled()) {
            anyline*  lines = buf[count].lines;
            vertinfo* ends0 = lines[0].ends;
            float ny = ends0[0].pos.x - ends0[1].pos.x;
            float nx = ends0[1].pos.y - ends0[0].pos.y;
            float c  = -nx * ends0[0].pos.x - ny * ends0[0].pos.y;

            // check winding
            bool ccw = true;
            for (int j = 1; ccw && j < buf[count].k; j++) {
                vertinfo* ends = lines[j].ends;
                for (int k = 0; ccw && k < 2; k++) {
                    float val = nx * ends[k].pos.x + ny * ends[k].pos.y + c;
                    if (val < -.1f) // allow small margin of error
                        ccw = false;
                }
            }

            if (ccw != (this->ctx_.cull_face() == CULL_BACK))
                continue; // back-face - skip
        }

        count++;
    }

    return count;
}

template<typename Uni, typename VAttr, typename VS, typename PS, typename Px>
void shader_obj_impl<Uni, VAttr, VS, PS, Px>::draw_clipped(
    int f, const rect& bound) {

    for (int i = 0; i < f; i++) {
        int       k    = clipped_[i].k;
        vertinfo* ctri = clipped_[i].verts;

        // draw by constructing trapezoids
        int left  = 1;
        int right = k-1;

        corner first_vert(ctri[0]);
        scanline top(int(ctri[0].pos.y), first_vert, first_vert);
        if (int(ctri[left].pos.y) == top.y)
            top.left = corner(ctri[left++]);
        else if (int(ctri[right].pos.y) == top.y)
            top.right = corner(ctri[right--]);

        // draw the trapezoids
        while (left <= right) {
            int ly = int(ctri[left].pos.y);
            int ry = int(ctri[right].pos.y);
            scanline btm;
            if (ly == ry) {
                btm = scanline(ly, corner(ctri[left]), corner(ctri[right]));
                left++;
                right--;
            } else if (ly < ry) {
                float t = float(ly - top.y) / float(ry - top.y);
                corner btm_r = corner(ctri[right]) * t
                             + top.right * (1.f - t);
                btm = scanline(ly, corner(ctri[left]), btm_r);
                left++;
            } else {
                float t = float(ry - top.y) / float(ly - top.y);
                corner btm_l = corner(ctri[left]) * t
                             + top.left * (1.f - t);
                btm = scanline(ry, btm_l, corner(ctri[right]));
                right--;
            }

            if (btm.y >= bound.ylo && top.y < bound.yhi)
                draw_trapezoid(top, btm, bound);

            top = btm;
        }

        draw_scanline(top, bound);
    }
}

template<typename Uni, typename VAttr, typename VS, typename PS, typename Px>
void shader_obj_impl<Uni, VAttr, VS, PS, Px>::screen_split_init(
    int f, const uint16_t* tris, rect* bounds, float* weights) {

    int inf       = std::numeric_limits<int>::max();
    int minus_inf = std::numeric_limits<int>::min();
    weights[0] = 0.f;
    bounds[0].xlo = inf;
    bounds[0].xhi = minus_inf;
    bounds[0].ylo = inf;
    bounds[0].yhi = minus_inf;
    for (int i = 0; i < f; i++) {
        weights[0] += clipped_[i].aabb.area();
        bounds[0].xlo = std::min(bounds[0].xlo, clipped_[i].aabb.xlo);
        bounds[0].xhi = std::max(bounds[0].xhi, clipped_[i].aabb.xhi);
        bounds[0].ylo = std::min(bounds[0].ylo, clipped_[i].aabb.ylo);
        bounds[0].yhi = std::max(bounds[0].yhi, clipped_[i].aabb.yhi);
    }
}

template<typename Uni, typename VAttr, typename VS, typename PS, typename Px>
bool shader_obj_impl<Uni, VAttr, VS, PS, Px>::screen_split1(
    int f, const uint16_t* tris, rect* bounds, float* weights,
    int i) {

    // find 'weightiest' rectangle to split
    int big_j = 0;
    int big_w = weights[0];
    for (int j = 1; j < i; j++) {
        int w = weights[j];
        if (w > big_w) {
            big_j = j;
            big_w = w;
        }
    }

    // big_j is the region to split
    rect& to_split = bounds[big_j];
    if (to_split.area() < 2500)
        return false;

    if (to_split.width() > to_split.height())
        return screen_split_horiz(f, tris, bounds, weights, i, big_j);
    else
        return screen_split_verti(f, tris, bounds, weights, i, big_j);

    return true;
}

template<typename Uni, typename VAttr, typename VS, typename PS, typename Px>
void shader_obj_impl<Uni, VAttr, VS, PS, Px>::draw_tris(
    int n, const vec3*     verts,   // verts[0..n-1]
    int f, const uint16_t* tris,    // tris[0..3f-1]
    const attr_arrays_t&   attrs) { // attrs[var][0..n-1]

    assert(this->ctx_.colorbuf() != NULL);
    assert(this->ctx_.depthbuf() != NULL);

    transform_verts(n, verts, attrs); // writes into vsout_
    if (this->ctx_.wireframe()) {
        draw_tris_wireframe(f, tris, attrs);
        return;
    }

    // fill mode
    f = transform_tris(f, tris);      // also culls, writes into clipped_
    int nthreads = this->ctx_.get_worker_pool().count();

    // partition for threads
    rect  bounds[nthreads];
    float weights[nthreads];
    screen_split_init(f, tris, bounds, weights);

    nsplits_ = nthreads - 1;
    for (int i = 1; i < nthreads; i++) {
        if (!screen_split1(f, tris, bounds, weights, i)) {
            nsplits_ = i - 1;
            break;
        }
    }

    // parallel draw
    int  nbounds = nsplits_ + 1;
    auto that    = this;
    this->ctx_.get_worker_pool().distribute_work([&](int thread) {
        if (thread < nbounds) {
            rect bound = bounds[thread];
            that->draw_clipped(f, bound);
        }
    });
}

template<typename Uni, typename VAttr, typename VS, typename PS, typename Px>
void shader_obj_impl<Uni, VAttr, VS, PS, Px>::draw_tris_wireframe(
    int f, const uint16_t* tris,    // tris[0..3f-1]
    const attr_arrays_t&   attrs) { // attrs[var][0..n-1]

    // vsout_ already filled by caller
    f = transform_tris_wireframe(f, tris); // also culls, writes into wftri_
    nsplits_ = 0;
    for (int i = 0; i < f; i++) {
        for (int j = 0; j < wftri_[i].k; j++)
            draw_screen_line(wftri_[i].lines[j]);
    }
}

// Nearly verbatim from
// https://en.wikipedia.org/wiki/Bresenham%27s_line_algorithm
// extended to interpolate and depth-test.
template<typename Uni, typename VAttr, typename VS, typename PS, typename Px>
void shader_obj_impl<Uni, VAttr, VS, PS, Px>::draw_screen_line(
    const anyline& line_in) {
    anyline line = line_in;

    // draw from top to bottom
    if (line.ends[0].pos.y > line.ends[1].pos.y) {
        std::swap(line.ends[0], line.ends[1]);
    } else if (line.ends[0].pos.y == line.ends[1].pos.y) {
        // or at least from left to right
        if (line.ends[0].pos.x > line.ends[1].pos.x)
            std::swap(line.ends[0], line.ends[1]);
    }

    // okay, draw
    int x0(line.ends[0].pos.x), y0(line.ends[0].pos.y);
    int x1(line.ends[1].pos.x), y1(line.ends[1].pos.y);

    int dx = abs(x1 - x0);
    int dy = abs(y1 - y0);
    int sx = x0 < x1 ? 1 : -1;
    int sy = y0 < y1 ? 1 : -1;
    int err = dx - dy;

    float  z    = line.ends[0].pos.z;
    qlin_t qlin = line.ends[0].qlin;
    vary_t vary = line.ends[0].vary;

    float  dz    = (line.ends[1].pos.z - z);
    qlin_t dqlin = (line.ends[1].qlin  - qlin);
    vary_t dvary = (line.ends[1].vary  - vary);

    float wrtx = 0.f;
    float wrty = 0.f;
    if (dx > dy)
        wrtx = 1.f / dx;
    else
        wrty = 1.f / dy;

    float  dz_x    = wrtx * dz;
    qlin_t dqlin_x = wrtx * dqlin;
    vary_t dvary_x = wrtx * dvary;

    float  dz_y    = wrty * dz;
    qlin_t dqlin_y = wrty * dqlin;
    vary_t dvary_y = wrty * dvary;

    int queue_pos = 0;
    typename simd<qlin_t>::type qlin_queue;
    typename simd<vary_t>::type vary_queue;
    int_4 x_queue;
    int_4 y_queue;
    float_4 z_queue;
    typename simd<Px>::type pxs;

    context<Uni, VAttr, Px>& ctx = this->ctx_;

    buffer<Px>*    tgt  = ctx.colorbuf();
    buffer<float>* zbuf = ctx.depthbuf();

    bool depth_test  = ctx.depth_test_enabled();
    bool depth_write = ctx.depth_write_enabled();
    bool color_write = ctx.color_write_enabled();

    for (;;) {
        if (!depth_test || z < zbuf->at(x0, y0)) {
            simd<qlin_t>::set(qlin_queue, queue_pos, qlin);
            simd<vary_t>::set(vary_queue, queue_pos, vary);
            x_queue.set(queue_pos, x0);
            y_queue.set(queue_pos, y0);
            z_queue.set(queue_pos, z);
            simd<Px>::set(pxs, queue_pos, tgt->at(x0, y0));
            ++queue_pos;

            if (queue_pos == 4) {
                auto ps_arg = (vary_queue, L_qlinear = qlin_queue.extract());
                ps_(this->ctx_.uniform(), ps_arg, x_queue, y_queue, z_queue, pxs);
                int xq0 = x_queue.get(0), yq0 = y_queue.get(0);
                int xq1 = x_queue.get(1), yq1 = y_queue.get(1);
                int xq2 = x_queue.get(2), yq2 = y_queue.get(2);
                int xq3 = x_queue.get(3), yq3 = y_queue.get(3);
                if (color_write) {
                    tgt->at(xq0, yq0) = simd<Px>::get(pxs, 0);
                    tgt->at(xq1, yq1) = simd<Px>::get(pxs, 1);
                    tgt->at(xq2, yq2) = simd<Px>::get(pxs, 2);
                    tgt->at(xq3, yq3) = simd<Px>::get(pxs, 3);
                }
                if (depth_write) {
                    zbuf->at(xq0, yq0) = z_queue.get(0);
                    zbuf->at(xq1, yq1) = z_queue.get(1);
                    zbuf->at(xq2, yq2) = z_queue.get(2);
                    zbuf->at(xq3, yq3) = z_queue.get(3);
                }
                queue_pos = 0;
            }
        }

        if (x0 == x1 && y0 == y1)
            break;

        int e2 = 2 * err;
        if (e2 > -dy) {
            err -= dy;
            x0  += sx;
            z   += dz_x;
            qlin = qlin + dqlin_x;
            vary = vary + dvary_x;
        }
        if (e2 < dx) {
            err += dx;
            y0  += sy;
            z   += dz_y;
            qlin = qlin + dqlin_y;
            vary = vary + dvary_y;
        }
    }

    if (queue_pos > 0) {
        auto ps_arg = (vary_queue, L_qlinear = qlin_queue.extract());
        ps_(this->ctx_.uniform(), ps_arg, x_queue, y_queue, z_queue, pxs);
        for (int i = 0; i < queue_pos; i++) {
            int x = x_queue.get(i), y = y_queue.get(i);
            if (color_write)
                tgt->at(x, y) = simd<Px>::get(pxs, i);
            if (depth_write)
                zbuf->at(x, y) = z_queue.get(i);
        }
    }
}

template<typename Uni, typename VAttr, typename VS, typename PS, typename Px>
void shader_obj_impl<Uni, VAttr, VS, PS, Px>::draw_splits(const Px& color) {
    int w = this->ctx_.width();
    int h = this->ctx_.height();
    for (int i = 0; i < nsplits_; i++) {
        if (splits_[i].xlo == splits_[i].xhi) {
            int x = splits_[i].xlo;
            int lo = std::max(0, splits_[i].ylo);
            int hi = std::min(h-1, splits_[i].yhi);
            for (int y = lo; y < hi; y++)
                this->ctx_.colorbuf()->at(x, y) = color;
        } else {
            int y = splits_[i].ylo;
            int lo = std::max(0, splits_[i].xlo);
            int hi = std::min(w-1, splits_[i].xhi);
            for (int x = lo; x < hi; x++)
                this->ctx_.colorbuf()->at(x, y) = color;
        }
    }
}

// shader_obj creation, needed for PX_MAKE_SHADER_OBJ()
template<typename Uni, typename VAttr, typename VS, typename PS, typename Px>
inline shader_obj_id make_shader_obj_(context<Uni, VAttr, Px>& ctx,
                                      const VS& vs, const PS& ps) {
    auto so = new shader_obj_impl<Uni, VAttr, VS, PS, Px>(ctx, vs, ps);
    std::unique_ptr<shader_obj<Uni, VAttr, Px> > so_uniq(so);
    return ctx.register_shader(std::move(so_uniq));
}

}      // namespace px
#endif // _PX_RASTERIZER_HH_
