#ifndef _PX_RECORD_HH_
#define _PX_RECORD_HH_
namespace px {

#define PX_REC_DEF_FIELD(name) static struct L_##name##_t {          \
    template<typename V>                                             \
    record<L_##name##_t, V, empty_rec_t> operator=(const V& val) {   \
        return record<L_##name##_t, V, empty_rec_t>(val, empty_rec); \
    }                                                                \
} L_##name; static inline void dummy_L_##name##_use() { (void)L_##name; }

template<typename L, typename U>
struct field;

template<typename L, typename U>
struct has_field;

template<typename L, typename V, typename R>
struct record;

static struct empty_rec_t {
    empty_rec_t() { }
    template<typename L, typename V, typename R>
    empty_rec_t(const record<L,V,R>& rec) { }
} empty_rec;

template<typename L, typename V, typename R>
struct record {
    V head_;
    R tail_;

    record() { } // V, R must be default-constructable

    record(const V& head, const R& tail) : head_(head), tail_(tail) {
        prevent_duplicates();
    }

    template<typename K, typename U, typename S>
    record(const record<K,U,S>& rec) : head_(rec[L()]), tail_(rec) {
        prevent_duplicates();
    }

    template<typename K>
    const typename field<K,record>::type& operator[](const K& k) const {
        return field<K,record>::ref(*this);
    }

    template<typename K>
    typename field<K,record>::type& operator[](const K& k) {
        return field<K,record>::ref(*this);
    }

    template<typename K, typename U, typename S>
    record& operator=(const record<K,U,S>& rec) {
        head_ = rec[L()];
        tail_ = rec;
        return *this;
    }

private:
    void prevent_duplicates() {
        static_assert(!has_field<L,R>::value, "duplicate field in record");
    }
};

template<typename L, typename K, typename V, typename R>
struct has_field<L, record<K,V,R> > {
    static const bool value = has_field<L,R>::value;
};

template<typename L, typename V, typename R>
struct has_field<L, record<L,V,R> > { static const bool value = true; };

template<typename L>
struct has_field<L,empty_rec_t> { static const bool value = false; };

template<typename L, typename V, typename R>
inline record<L,V,R> operator,(const R& tail, const record<L,V,empty_rec_t>& head) {
    return record<L,V,R>(head.head_, tail);
}

template<typename L, typename V, typename R>
struct field<L, record<L,V,R> > {
    typedef V type;
    static const type& ref(const record<L,V,R>& r) { return r.head_; }
    static type& ref(record<L,V,R>& r) { return r.head_; }
};

template<typename L, typename K, typename V, typename R>
struct field<L, record<K,V,R> > {
    typedef typename field<L,R>::type type;
    static const type& ref(const record<K,V,R>& r) {
        return field<L,R>::ref(r.tail_);
    }
    static type& ref(record<K,V,R>& r) {
        return field<L,R>::ref(r.tail_);
    }
};

template<typename L, typename U>
struct field_split;

template<typename L>
struct field_split<L, empty_rec_t> {
    typedef empty_rec_t full_t;
    typedef empty_rec_t focus_t;
    typedef empty_rec_t nonfocus_t;

    static focus_t focus(const full_t& full) { return empty_rec; }
    static nonfocus_t nonfocus(const full_t& full) { return empty_rec; }
};

template<typename L, typename V, typename R>
struct field_split<L, record<L,V,R> > {
    typedef record<L,V,R> full_t;
    typedef V focus_t;
    typedef R nonfocus_t;

    static focus_t focus(const full_t& full) { return full.head_; }
    static nonfocus_t nonfocus(const full_t& full) { return full.tail_; }
};

template<typename L, typename K, typename V, typename R>
struct field_split<L, record<K,V,R> > {
    typedef record<K,V,R> full_t;
    typedef typename field_split<L,R>::focus_t focus_t;
    typedef record<K, V, typename field_split<L,R>::nonfocus_t> nonfocus_t;

    static focus_t focus(const full_t& full) {
        return field_split<L, R>::focus(full.tail_);
    }
    static nonfocus_t nonfocus(const full_t& full) {
        return (field_split<L, R>::nonfocus(full.tail_), K() = full.head_);
    }
};

/*
 * Addition, subtraction, and scalar multiplication
 * distribute across records.
 */

inline empty_rec_t operator-(const empty_rec_t& a) { return empty_rec; }

inline empty_rec_t operator+(const empty_rec_t& a, const empty_rec_t& b) {
    return empty_rec;
}

inline empty_rec_t operator-(const empty_rec_t& a, const empty_rec_t& b) {
    return empty_rec;
}

template<typename U>
inline empty_rec_t operator*(const U& a, const empty_rec_t& b) {
    return empty_rec;
}

template<typename L, typename V, typename R>
inline record<L,V,R> operator-(const record<L,V,R>& a) {
    return record<L,V,R>(-a.head_, -a.tail_);
}

template<typename L, typename V, typename R>
inline record<L,V,R> operator+(const record<L,V,R>& a, const record<L,V,R>& b) {
    return record<L,V,R>(a.head_ + b.head_, a.tail_ + b.tail_);
}

template<typename L, typename V, typename R>
inline record<L,V,R> operator-(const record<L,V,R>& a, const record<L,V,R>& b) {
    return record<L,V,R>(a.head_ - b.head_, a.tail_ - b.tail_);
}

template<typename U, typename L, typename V, typename R>
inline record<L,V,R> operator*(const U& a, const record<L,V,R>& b) {
    return record<L,V,R>(a * b.head_, a * b.tail_);
}

/* Wrap each record field in a pointer, for arrays. */
template<typename U>
struct rec_arrays;

template<>
struct rec_arrays<empty_rec_t> {
    typedef empty_rec_t type;
    static empty_rec_t index(const type& arrays, int i) { return empty_rec; }
};

template<typename L, typename V, typename R>
struct rec_arrays<record<L,V,R> > {
    typedef record<L, const V*, typename rec_arrays<R>::type> type;
    static record<L,V,R> index(const type& arrays, int i) {
        return record<L,V,R>(arrays.head_[i],
                             rec_arrays<R>::index(arrays.tail_, i));
    }
};

}      // namespace px
#endif // _PX_RECORD_HH_
