#ifndef _PX_SIMD_HH_
#define _PX_SIMD_HH_

#include <cstdint>
#include <xmmintrin.h>
#include <emmintrin.h>

namespace px {

// 4-at-a-time version of bool (used for comparisons and whatnot)
struct mask_4 {
    __m128 val;

    // construction
    mask_4() { }
    mask_4(const __m128& v) : val(v) { }
    mask_4(const mask_4& v) : val(v.val) { }
    mask_4(bool a, bool b, bool c, bool d) : val(convert(a, b, c, d)) { }

    explicit mask_4(bool x) : val(convert(x, x, x, x)) { }

    // assignment
    mask_4& operator=(const mask_4& m) { val = m.val; return *this; }

    // access by index
    bool get(int i) const {
        __m128i vi = (__m128i)val;
        return ((const int32_t*) &vi)[i] != 0;
    }
    void set(int i, bool x) {
        bool xs[4] = { false, false, false, false };
        xs[i] = x;
        __m128 v = convert(xs[0], xs[1], xs[2], xs[3]);
        if (x)
            val = _mm_or_ps(val, v);
        else
            val = _mm_andnot_ps(v, val);
    }

    // logical operations
    mask_4 operator~() const
        { return mask_4(_mm_andnot_ps(val, mask_4(true).val)); }
    mask_4 operator&(const mask_4& m) const
        { return mask_4(_mm_and_ps(val, m.val)); }
    mask_4 operator|(const mask_4& m) const
        { return mask_4(_mm_or_ps(val, m.val)); }
    mask_4 andnot(const mask_4& m) const
        { return mask_4(_mm_andnot_ps(m.val, val)); }

    // destructive logical operations
    mask_4& complement_d() { return *this = ~(*this); }
    mask_4& operator&=(const mask_4& m) { return *this = *this & m; }
    mask_4& operator|=(const mask_4& m) { return *this = *this | m; }
    mask_4& andnot_d(const mask_4& m) { return *this = andnot(m); }

    // aggregate logical operations
    bool all()  const { return movemask() == 0xffff; } // 16-bit mask
    bool any()  const { return movemask() != 0; }
    bool none() const { return movemask() == 0; }

private:
    int movemask() const { return _mm_movemask_epi8((__m128i)val); }

    static float  convert1(bool x) { return x ? 1.f : -1.f; }
    static __m128 convert (bool a, bool b, bool c, bool d) {
        __m128 v;
        v = _mm_set_ps(convert1(d), convert1(c), convert1(b), convert1(a));
        return _mm_cmple_ps(_mm_setzero_ps(), v);
    }
};

// 4-at-a-time version of float
struct float_4 {
    __m128 val;

    // construction
    float_4() { }
    float_4(const __m128& v) : val(v) { }
    float_4(const float_4& s) : val(s.val) { }
    float_4(float a, float b, float c, float d)
        : val(_mm_set_ps(d, c, b, a)) { }

    explicit float_4(float a) : val(_mm_set1_ps(a)) { }

    // assignment
    float_4& operator=(const float_4& v) { val = v.val; return *this; }

    // access by index
    float& at(int i) { return ((float*)&val)[i]; }
    const float& at(int i) const { return ((float*)&val)[i]; }

    float get(int i) const { return at(i); }
    void  set(int i, float x) { at(i) = x; }

    // arithmetic
    float_4 operator-() const { return float_4(0.f) - *this; }

    float_4 operator+(const float_4& x) const {
        return float_4(_mm_add_ps(val, x.val));
    }

    float_4 operator-(const float_4& x) const {
        return float_4(_mm_sub_ps(val, x.val));
    }

    float_4 operator*(const float_4& x) const {
        return float_4(_mm_mul_ps(val, x.val));
    }

    float_4 operator/(const float_4& x) const {
        return float_4(_mm_div_ps(val, x.val));
    }

    // destructive arithmetic
    float_4& operator+=(const float_4& x) {
        val = _mm_add_ps(val, x.val); return *this;
    }

    float_4& operator-=(const float_4& x) {
        val = _mm_sub_ps(val, x.val); return *this;
    }

    float_4& operator*=(const float_4& x) {
        val = _mm_mul_ps(val, x.val); return *this;
    }

    float_4& operator/=(const float_4& x) {
        val = _mm_div_ps(val, x.val); return *this;
    }

    // functions
    float_4 recip() const { return float_4(_mm_rcp_ps(val)); }
    float_4 sqrt()  const { return float_4(_mm_sqrt_ps(val)); }
    float_4 rsqrt() const { return float_4(_mm_rsqrt_ps(val)); }
    float_4 min(const float_4& x) { return float_4(_mm_min_ps(val, x.val)); }
    float_4 max(const float_4& x) { return float_4(_mm_max_ps(val, x.val)); }
    float_4 clamp(const float_4& low, const float_4& high) {
        return min(high).max(low);
    }

    // destructive functions
    float_4& recip_d() { val = _mm_rcp_ps(val);   return *this; }
    float_4& sqrt_d()  { val = _mm_sqrt_ps(val);  return *this; }
    float_4& rsqrt_d() { val = _mm_rsqrt_ps(val); return *this; }
    float_4& min_d(const float_4& x) { return *this = min(x); }
    float_4& max_d(const float_4& x) { return *this = max(x); }
    float_4& clamp_d(const float_4& low, const float_4& high) {
        return *this = clamp(low, high);
    }

    // masking
    float_4 mask(const mask_4& m) const {
        return float_4(_mm_and_ps(val, m.val));
    }
    float_4& mask_d(const mask_4& m) {
        val = _mm_and_ps(val, m.val);
        return *this;
    }

    // comparison
    mask_4 operator==(const float_4& x) const {
        return mask_4(_mm_cmpeq_ps(val, x.val));
    }

    mask_4 operator!=(const float_4& x) const {
        return mask_4(_mm_cmpneq_ps(val, x.val));
    }

    mask_4 operator<(const float_4& x) const {
        return mask_4(_mm_cmplt_ps(val, x.val));
    }

    mask_4 operator<=(const float_4& x) const {
        return mask_4(_mm_cmple_ps(val, x.val));
    }

    mask_4 operator>(const float_4& x) const {
        return mask_4(_mm_cmpgt_ps(val, x.val));
    }

    mask_4 operator>=(const float_4& x) const {
        return mask_4(_mm_cmpge_ps(val, x.val));
    }
};

// 4-at-a-time version of int32_t
struct int_4 {
    __m128i val;

    // construction
    int_4() { }
    int_4(const __m128i& v) : val(v) { }
    int_4(const int_4& s) : val(s.val) { }
    int_4(int32_t a, int32_t b, int32_t c, int32_t d)
        : val(_mm_set_epi32(d, c, b, a)) { }

    explicit int_4(int32_t a) : val(_mm_set1_epi32(a)) { }
    explicit int_4(const float_4& a) : val(_mm_cvtps_epi32(a.val)) { }

    // assignment
    int_4& operator=(const int_4& v) { val = v.val; return *this; }

    // conversion
    explicit operator float_4() const { return float_4(_mm_cvtepi32_ps(val)); }

    // access by index
    int32_t&       at(int i) { return ((int32_t*)&val)[i]; }
    const int32_t& at(int i) const { return ((int32_t*)&val)[i]; }

    int32_t get(int i) const { return at(i); }
    void    set(int i, int32_t x) { at(i) = x; }

    // arithmetic
    int_4 operator-() const { return int_4(0u) - *this; }

    int_4 operator+(const int_4& x) const {
        return int_4(_mm_add_epi32(val, x.val));
    }

    int_4 operator-(const int_4& x) const {
        return int_4(_mm_sub_epi32(val, x.val));
    }

    int_4 operator*(const int_4& x) const {
        // http://stackoverflow.com/questions/10500766/
        //   sse-multiplication-of-4-32-bit-integers
        __m128i tmp1 = _mm_mul_epu32(val, x.val); // mul 2,0
        __m128i tmp2 = _mm_mul_epu32(
            _mm_srli_si128(val,   4),
            _mm_srli_si128(x.val, 4)); // mul 3,1
        return int_4(_mm_unpacklo_epi32(
            _mm_shuffle_epi32(tmp1, _MM_SHUFFLE (0,0,2,0)),
            _mm_shuffle_epi32(tmp2, _MM_SHUFFLE (0,0,2,0))));
    }

    // destructive arithmetic
    int_4& operator+=(const int_4& x) {
        val = _mm_add_epi32(val, x.val); return *this;
    }

    int_4& operator-=(const int_4& x) {
        val = _mm_sub_epi32(val, x.val); return *this;
    }

    // masking
    int_4 mask(const mask_4& m) const {
        return int_4(_mm_and_si128(val, (__m128i)m.val));
    }
    int_4& mask_d(const mask_4& m) {
        val = _mm_and_si128(val, (__m128i)m.val);
        return *this;
    }

    // comparison
    mask_4 operator==(const int_4& x) const {
        return mask_4((__m128)_mm_cmpeq_epi32(val, x.val));
    }

    mask_4 operator!=(const int_4& x) const {
        return ~(*this == x);
    }

    mask_4 operator<(const int_4& x) const {
        return mask_4((__m128)_mm_cmplt_epi32(val, x.val));
    }

    mask_4 operator<=(const int_4& x) const {
        return ~(*this > x);
    }

    mask_4 operator>(const int_4& x) const {
        return mask_4((__m128)_mm_cmpgt_epi32(val, x.val));
    }

    mask_4 operator>=(const int_4& x) const {
        return ~(*this < x);
    }

    // bitwise ops
    int_4 operator~() const {
        return int_4(_mm_andnot_si128(val, (__m128i)mask_4(true).val));
    }
    int_4 operator&(const int_4& x) const {
        return int_4(_mm_and_si128(val, x.val));
    }
    int_4 operator|(const int_4& x) const {
        return int_4(_mm_or_si128(val, x.val));
    }
    int_4 operator^(const int_4& x) const {
        return int_4(_mm_xor_si128(val, x.val));
    }

    // destructive bitwise ops
    int_4& complement_d() { return *this = ~(*this); }
    int_4& operator&=(const int_4& x) { return *this = *this & x; }
    int_4& operator|=(const int_4& x) { return *this = *this | x; }
    int_4& operator^=(const int_4& x) { return *this = *this ^ x; }

    // shifts (logical)
    int_4 operator<<(int count) const {
        return int_4(_mm_slli_epi32(val, count));
    }
    int_4 operator>>(int count) const {
        return int_4(_mm_srli_epi32(val, count));
    }

    // destructive shifts
    int_4& operator<<=(int count) { return *this = *this << count; }
    int_4& operator>>=(int count) { return *this = *this >> count; }
};

// 4-at-a-time version of uint32_t
struct uint_4 {
    __m128i val;

    // construction
    uint_4() { }
    uint_4(const __m128i& v) : val(v) { }
    uint_4(const uint_4& s) : val(s.val) { }
    uint_4(uint32_t a, uint32_t b, uint32_t c, uint32_t d)
        : val(_mm_set_epi32(d, c, b, a)) { }

    explicit uint_4(uint32_t a) : val(_mm_set1_epi32(a)) { }
    explicit uint_4(const float_4& a) : val(_mm_cvtps_epi32(a.val)) { }

    // assignment
    uint_4& operator=(const uint_4& v) { val = v.val; return *this; }

    // conversion
    explicit operator float_4() const { return float_4(_mm_cvtepi32_ps(val)); }

    // access by index
    uint32_t&       at(uint i) { return ((uint32_t*)&val)[i]; }
    const uint32_t& at(uint i) const { return ((uint32_t*)&val)[i]; }

    uint32_t get(uint i) const { return at(i); }
    void     set(uint i, uint32_t x) { at(i) = x; }

    // arithmetic
    uint_4 operator-() const { return uint_4(0u) - *this; }

    uint_4 operator+(const uint_4& x) const {
        return uint_4(_mm_add_epi32(val, x.val));
    }

    uint_4 operator-(const uint_4& x) const {
        return uint_4(_mm_sub_epi32(val, x.val));
    }

    // destructive arithmetic
    uint_4& operator+=(const uint_4& x) {
        val = _mm_add_epi32(val, x.val); return *this;
    }

    uint_4& operator-=(const uint_4& x) {
        val = _mm_sub_epi32(val, x.val); return *this;
    }

    // masking
    uint_4 mask(const mask_4& m) const {
        return uint_4(_mm_and_si128(val, (__m128i)m.val));
    }
    uint_4& mask_d(const mask_4& m) {
        val = _mm_and_si128(val, (__m128i)m.val);
        return *this;
    }

    // comparison
    mask_4 operator==(const uint_4& x) const {
        return mask_4((__m128)_mm_cmpeq_epi32(val, x.val));
    }

    mask_4 operator!=(const uint_4& x) const {
        return ~(*this == x);
    }

    mask_4 operator<(const uint_4& x) const {
        return mask_4((__m128)_mm_cmplt_epi32(val, x.val));
    }

    mask_4 operator<=(const uint_4& x) const {
        return ~(*this > x);
    }

    mask_4 operator>(const uint_4& x) const {
        return mask_4((__m128)_mm_cmpgt_epi32(val, x.val));
    }

    mask_4 operator>=(const uint_4& x) const {
        return ~(*this < x);
    }

    // bitwise ops
    uint_4 operator~() const {
        return uint_4(_mm_andnot_si128(val, (__m128i)mask_4(true).val));
    }
    uint_4 operator&(const uint_4& x) const {
        return uint_4(_mm_and_si128(val, x.val));
    }
    uint_4 operator|(const uint_4& x) const {
        return uint_4(_mm_or_si128(val, x.val));
    }
    uint_4 operator^(const uint_4& x) const {
        return uint_4(_mm_xor_si128(val, x.val));
    }

    // destructive bitwise ops
    uint_4& complement_d() { return *this = ~(*this); }
    uint_4& operator&=(const uint_4& x) { return *this = *this & x; }
    uint_4& operator|=(const uint_4& x) { return *this = *this | x; }
    uint_4& operator^=(const uint_4& x) { return *this = *this ^ x; }

    // shifts (logical)
    uint_4 operator<<(uint count) const {
        return uint_4(_mm_slli_epi32(val, count));
    }
    uint_4 operator>>(uint count) const {
        return uint_4(_mm_srli_epi32(val, count));
    }

    // destructive shifts
    uint_4& operator<<=(uint count) { return *this = *this << count; }
    uint_4& operator>>=(uint count) { return *this = *this >> count; }
};

// generic 4-at-a-time version of arbitrary types
template<typename U>
struct quadruple {
    U vals_[4];

    // construction
    quadruple() { }
    quadruple(const quadruple<U>& q) {
        for (int i = 0; i < 4; i++)
            vals_[i] = q.vals_[i];
    }

    quadruple(const U& a, const U& b, const U& c, const U& d) {
        vals_[0] = a;
        vals_[1] = b;
        vals_[2] = c;
        vals_[3] = d;
    }

    explicit quadruple(const U& x) {
        for (int i = 0; i < 4; i++)
            vals_[i] = x;
    }

    // assignment
    quadruple<U>& operator=(const quadruple<U>& q) {
        vals_[0] = q.vals_[0];
        vals_[1] = q.vals_[1];
        vals_[2] = q.vals_[2];
        vals_[3] = q.vals_[3];
        return *this;
    }

    // access
    U& at(int i) { return vals_[i]; }
    const U& at(int i) const { return vals_[i]; }

    U get(int i) const { return vals_[i]; }
    void set(int i, const U& val) { vals_[i] = val; }
};

// simd-ify a type
template<typename U>
struct simd {
    typedef quadruple<U> type;
    static type make(const U& a, const U& b, const U& c, const U& d) {
        return quadruple<U>(a, b, c, d);
    }
    static U    get(const type& q, int i) { return q.get(i); }
    static void set(type& q, int i, const U& v) { q.set(i, v); }
    static const type choose(
        const mask_4& m, const type& then, const type& els) {
        return quadruple<U>(
            m.get(0) ? then.get(0) : els.get(0),
            m.get(1) ? then.get(1) : els.get(1),
            m.get(2) ? then.get(2) : els.get(2),
            m.get(3) ? then.get(3) : els.get(3));
    }
};

template<>
struct simd<float> {
    typedef float_4 type;
    static type make(float a, float b, float c, float d) {
        return float_4(a, b, c, d);
    }
    static float get(const type& f4, int i) { return f4.get(i); }
    static void  set(type& f4, int i, float v) { f4.set(i, v); }
    static type choose(
        const mask_4& m, const float_4& then, const float_4& els) {
        return float_4(_mm_or_ps(
            _mm_and_ps(m.val, then.val),
            _mm_andnot_ps(m.val, els.val)));
    }
};

template<>
struct simd<bool> {
    typedef mask_4 type;
    static type make(bool a, bool b, bool c, bool d) {
        return mask_4(a, b, c, d);
    }
    static bool get(const type& b4, int i) { return b4.get(i); }
    static void set(type& b4, int i, bool b) { b4.set(i, b); }
    static type choose(
        const mask_4& m, const mask_4& then, const mask_4& els) {
        return (then & m) | (els.andnot(m));
    }
};

template<>
struct simd<int32_t> {
    typedef int_4 type;
    static type make(int32_t a, int32_t b, int32_t c, int32_t d) {
        return int_4(a, b, c, d);
    }
    static int32_t get(const type& q, int i) { return q.get(i); }
    static void set(type& q, int i, int32_t x) { q.set(i, x); }
    static type choose(const mask_4& m, const type& then, const type& els) {
        return then.mask(m) | els.mask(~m);
    }
};

template<>
struct simd<uint32_t> {
    typedef uint_4 type;
    static type make(uint32_t a, uint32_t b, uint32_t c, uint32_t d) {
        return uint_4(a, b, c, d);
    }
    static uint32_t get(const type& q, int i) { return q.get(i); }
    static void set(type& q, int i, uint32_t x) { q.set(i, x); }
    static type choose(const mask_4& m, const type& then, const type& els) {
        return then.mask(m) | els.mask(~m);
    }
};

// construct by duplicating a single value
template<typename U>
inline typename simd<U>::type simd_make1(const U& val) {
    return simd<U>::make(val, val, val, val);
}

}      // namespace px
#endif // _PX_SIMD_HH_
