#ifndef _PX_WORKER_HH_
#define _PX_WORKER_HH_

#include <pthread.h>
#include <semaphore.h>

namespace px {

class worker_pool {
private:
    // private because use is error-prone (see worker_pool::distribute_work())
    class worker {
        bool      wasinit_, running_;
        pthread_t thread_;
        sem_t     req_, done_;
        const std::function<void()>* volatile work_;

    public:
        worker() : wasinit_(false), running_(false), work_(NULL) { }
        ~worker() {
            assert(!running_);
            if (wasinit_) {
                sem_destroy(&req_);
                sem_destroy(&done_);
            }
        }

        bool init() {
            assert(!wasinit_);
            if (sem_init(&req_,  0, 0) < 0)
                return false;
            if (sem_init(&done_, 0, 0) < 0) {
                sem_destroy(&req_);
                return false;
            }
            wasinit_ = true;
            return true;
        }

        bool start() {
            assert(!running_);
            void* (*tmain)(void*) = (void*(*)(void*))&thread_main;
            if (pthread_create(&thread_, NULL, tmain, this) != 0)
                return false;

            running_ = true;
            return true;
        }

        void stop() {
            assert(running_);
            work_ = NULL;
            sem_post(&req_);
            pthread_join(thread_, NULL);
            running_ = false;
        }

        void send(const std::function<void()>& fn) {
            assert(&fn != NULL);
            work_ = &fn;
            int rc = sem_post(&req_);
            assert(rc == 0); (void)rc;
        }

        void wait() {
            int rc;
            do rc = sem_wait(&done_); while (rc == EINTR);
            assert(rc == 0);
            work_ = NULL;
        }

    private:
        static void* thread_main(worker* w) {
            for (;;) {
                int rc;
                do rc = sem_wait(&w->req_); while (rc == EINTR);
                assert(rc == 0);

                if (w->work_ == NULL)
                    return NULL;

                (*w->work_)();
                rc = sem_post(&w->done_);
                assert(rc == 0);
            }
        }
    };


    int n_;
    worker* workers_;

public:
    worker_pool(int n) : n_(n-1), workers_(NULL) {
        assert(n >= 1);
    }

    ~worker_pool() {
        if (workers_ == NULL)
            return;
        for (int i = 0; i < n_; i++)
            workers_[i].stop();
        delete[] workers_;
    }

    int  count()   { return n_ + 1; }
    bool wasinit() { return workers_ != NULL; }

    bool init() {
        assert(workers_ == NULL);
        workers_ = new worker[n_];
        for (int i = 0; i < n_; i++) {
            if (workers_[i].init() && workers_[i].start())
                continue; // success
            // failure
            for (int j = 0; j < i; j++)
                workers_[j].stop();
            delete[] workers_;
            workers_ = NULL;
            return false;
        }
        return true;
    }

    // runs each_index(i) for i = 0, ..., count() - 1
    void distribute_work(std::function<void(int)> each_index) {
        assert(workers_ != NULL);

        // avoid temporaries stepping on each other's toes
        std::function<void()> each[n_];
        for (int i = 0; i < n_; i++)
            each[i] = [=]() { each_index(i); }; 

        // run each function
        for (int i = 0; i < n_; i++)
            workers_[i].send(each[i]);

        each_index(n_); // final call is in calling thread

        // wait for completion
        for (int i = 0; i < n_; i++)
            workers_[i].wait();
    }
};

}      // namespace px
#endif // _PX_WORKER_HH_
