#include "sdlutil.hh"
#include <cstdlib>
#include <cstdio>

#ifndef NO_SDL_IMAGE
#include "SDL_image.h"
#else
#include <png.h>
#endif

static char err[1024]; // NOT THREAD SAFE

bool sdlutil_init(const char** error) {
    *error = err;
    err[0] = '\0';

    // Initialize SDL
    if (SDL_Init(SDL_INIT_VIDEO) < 0) {
        snprintf(err, sizeof (err), "SDL_Init: %s\n", SDL_GetError());
        return false;
    }

#ifndef NO_SDL_IMAGE
    // Initialize SDL_image
    int img_flags = IMG_INIT_JPG | IMG_INIT_PNG | IMG_INIT_TIF;
    if (IMG_Init(img_flags) != img_flags) {
        snprintf(err, sizeof (err), "IMG_Init: %s\n", IMG_GetError());
        SDL_Quit();
        return false;
    }
#endif

    // Initialize SDL_ttf
#ifndef NO_SDL_TTF
    if (TTF_Init() < 0) {
        snprintf(err, sizeof (err), "TTF_Init: %s\n", TTF_GetError());
#ifndef NO_SDL_IMAGE
        IMG_Quit();
#endif
        SDL_Quit();
        return false;
    }
#endif

    atexit(&SDL_Quit);
#ifndef NO_SDL_IMAGE
    atexit(&IMG_Quit);
#endif
#ifndef NO_SDL_TTF
    atexit(&TTF_Quit);
#endif
    return true;
}

std::unique_ptr<px::buffer<px::vec3> > load_rgb(
    const char* filename, const char **error) {

    using px::buffer;
    using px::vec3;

#ifndef NO_SDL_IMAGE

    *error = err;
    err[0] = '\0';

    unique_surface_ptr surf(IMG_Load(filename));
    if (surf == NULL) {
        snprintf(err, sizeof (err), "IMG_Load: %s\n", IMG_GetError());
        return NULL;
    }

    const SDL_PixelFormat* fmt = surf->format;
    int Bpp = fmt->BytesPerPixel;
    if (Bpp > 4) {
        snprintf(err, sizeof (err),
                 "%s: too many bytes per pixel (%d)\n",
                 filename, Bpp);
        return NULL;
    }

    buffer<vec3>* tex   = new buffer<vec3>(surf->w, surf->h);
    float         scale = 1.f / 255.f;
    for (int y = 0, inv_y = surf->h - 1; y < surf->h; y++, inv_y--) {
        Uint8* p = (Uint8*)surf->pixels + y * surf->pitch;
        if (Bpp == 1) {
            for (int x = 0; x < surf->w; x++) {
                SDL_Color *c = &fmt->palette->colors[*p++];
                tex->at(x, inv_y) = scale * vec3(c->r, c->g, c->b);
            }
            continue;
        }

        // Bpp > 1
        for (int x = 0; x < surf->w; x++) {
            Uint32 px_bytes = 0;
            for (int i = Bpp-1; i >= 0; i--) {
                px_bytes <<= 8;
                px_bytes  |= p[i];
            }
            p += Bpp;
            tex->at(x, inv_y) = scale * vec3(
                ((px_bytes & fmt->Rmask) >> fmt->Rshift) << fmt->Rloss,
                ((px_bytes & fmt->Gmask) >> fmt->Gshift) << fmt->Gloss,
                ((px_bytes & fmt->Bmask) >> fmt->Bshift) << fmt->Bloss);
        }
    }

    return std::unique_ptr<buffer<vec3> >(tex);

#else
    // SDL_NO_IMAGE is defined
    // http://stackoverflow.com/questions/2703255/
    //     need-help-using-libpng-to-read-an-image
    png_structp png_ptr;
    png_infop info_ptr;
    FILE *fp;

    if ((fp = fopen(filename, "rb")) == NULL) {
        snprintf(err, sizeof (err), "%s: %s", filename, strerror(errno));
        return false;
    }

    png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);

    if (png_ptr == NULL) {
        snprintf(err, sizeof (err),
                 "%s: png_create_read_struct() failed\n", filename);
        fclose(fp);
        return NULL;
    }

    info_ptr = png_create_info_struct(png_ptr);
    if (info_ptr == NULL) {
        snprintf(err, sizeof (err),
                 "%s: png_create_info_struct() failed\n", filename);
        fclose(fp);
        png_destroy_read_struct(&png_ptr, NULL, NULL);
        return NULL;
    }

    if (setjmp(png_jmpbuf(png_ptr))) {
        snprintf(err, sizeof (err), "%s: loading png failed\n", filename);
        png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
        fclose(fp);
        return NULL;
    }

    png_init_io(png_ptr, fp);

    png_read_png(png_ptr, info_ptr,
                   PNG_TRANSFORM_STRIP_16
                 | PNG_TRANSFORM_SWAP_ALPHA
                 | PNG_TRANSFORM_EXPAND,
                 NULL);

    png_uint_32 width = png_get_image_width(png_ptr, info_ptr);
    png_uint_32 height = png_get_image_height(png_ptr, info_ptr);

    png_uint_32 channels   = png_get_channels(png_ptr, info_ptr);
    //png_uint_32 bitdepth   = png_get_bit_depth(png_ptr, info_ptr);
    //png_uint_32 color_type = png_get_color_type(png_ptr, info_ptr);

    png_bytepp row_pointers;
    row_pointers = png_get_rows(png_ptr, info_ptr);

    buffer<vec3>* tex   = new buffer<vec3>(width, height);
    float         scale = 1.f / 255.f;

    for (unsigned y = 0, inv_y = height-1; y < height; y++, inv_y--) {
        uint8_t* row = row_pointers[y];
        for (unsigned x = 0; x < width; x++) {
            // probably really broken
            if (channels == 4)
                row++; // ignore alpha
            uint8_t  r = *row++;
            uint8_t  g = *row++;
            uint8_t  b = *row++;
            tex->at(x, inv_y) = scale * vec3(r, g, b);
        }
    }

    return std::unique_ptr<buffer<vec3> >(tex);

#endif

}
