#include "px/px.hh"
#include "sdlutil.hh"
#include "mesh.hh"
#include "framerate.hh"

#include <cstdio>
#include <cstdlib>
#include <cstdint>
#include <cmath>
#include <cassert>

using namespace px;

PX_REC_DEF_FIELD(texcoord);
PX_REC_DEF_FIELD(normal);
PX_REC_DEF_FIELD(projection);
PX_REC_DEF_FIELD(modelview);
PX_REC_DEF_FIELD(light_dir);
PX_REC_DEF_FIELD(texture);

#define DEFAULT_NTHREADS      1
#define DEFAULT_SCREEN_FLAGS  (SDL_HWSURFACE)
#define DEFAULT_SCREEN_WIDTH  640
#define DEFAULT_SCREEN_HEIGHT 400
#define DEFAULT_FONT_FILE     "data/font.ttf"
#define DEFAULT_OBJ_FILE      "data/model.obj"
#define DEFAULT_TEX_FILE      "data/texture.png"

struct options {
    int nthreads;
    int screen_flags;
    int screen_width;
    int screen_height;
    const char* font_file;
    const char* obj_file;
    const char* tex_file;
};

options parse_options(int argc, char* argv[]);

// context/shader_obj types
typedef record<L_projection_t, mat4,
        record<L_modelview_t,  mat4,
        record<L_light_dir_t,  vec3,
        record<L_texture_t,    buffer<vec3>*, empty_rec_t> > > >
    Uni;

typedef record<L_texcoord_t,   vec2,
        record<L_normal_t,     vec3, empty_rec_t> >
    VAttr;

typedef Uint32 Px;

// shared parts of pixel shader
inline uint_4 convert_colors(vec3_4 c);

template<typename Vary>
inline void color_and_light(const Uni& uni, const Vary& vary,
                            vec3_4& c, float_4& light_val) {
    // sample texture to get color
    c = uni[L_texture]->sample(vary[L_qlinear][L_texcoord]);

    // compute light contribution (white light)
    float_4 zero(0.f), one(1.f);
    const vec3_4 reverse_light_dir(uni[L_light_dir]);
    const vec3_4 normal = vary[L_normal].normalize();
    float_4 light_dot = reverse_light_dir.dot(normal);
    light_val = float_4(.2f) + light_dot.max(zero);
}

int main(int argc, char *argv[]) {
    struct options opt = parse_options(argc, argv);

    // initialize px
    px_init();

    // initialize sdl
    const char* errmsg;
    if (!sdlutil_init(&errmsg)) {
        fprintf(stderr, "error: %s\n", errmsg);
        exit(1);
    }

    // open video
    unique_surface_ptr screen(SDL_SetVideoMode(
        opt.screen_width, opt.screen_height, 32, opt.screen_flags));
    if (screen == NULL) {
        fprintf(stderr, "error: SDL_SetVideoMode: %s\n", SDL_GetError());
        exit(1);
    }

    if (screen->format->BytesPerPixel != 4) {
        fprintf(stderr, "fatal: not 4 bytes per pixel!\n");
        exit(1);
    }

    // load a font
#ifndef NO_SDL_TTF
    unique_font_ptr font(TTF_OpenFont(opt.font_file, 24));
    if (font == NULL) {
        fprintf(stderr, "error: TTF_OpenFont: %s\n", TTF_GetError());
        exit(1);
    }

    TTF_SetFontStyle(font.get(), TTF_STYLE_BOLD);
#else
    uint32_t last_fps_print = 0;
#endif

    // load model
    mesh mod;
    if (!mod.load(opt.obj_file)) {
        fprintf(stderr, "error: %s\n", mod.error);
        exit(1);
    }

    // find a natural 'centre'
    vec3 mod_center = vec3::zero();
    float mod_maxdim = 0.f;
    for (int i = 0; i < mod.nverts; i++)
        mod_center += mod.verts[i];

    mod_center *= 1.f / mod.nverts;

    // find max distance (along a single dimension) from centre
    for (int i = 0; i < mod.nverts; i++) {
        const vec3& v = mod.verts[i];
        vec3  d     = v - mod_center;
        float dim   = std::max(fabs(d.x), std::max(fabs(d.y), fabs(d.z)));
        mod_maxdim  = std::max(mod_maxdim, dim);
    }

    // load texture
    std::unique_ptr<buffer<vec3> > texture = load_rgb(opt.tex_file, &errmsg);
    if (texture == NULL) {
        fprintf(stderr, "error: %s\n", errmsg);
        exit(1);
    }

    // Set up renderer
    buffer<Px> target(opt.screen_width, opt.screen_height,
                      (simd<Px>::type*)screen->pixels);

    Uni initial_uni = (L_projection = mat4::identity(),
                       L_modelview  = mat4::identity(),
                       L_light_dir  = vec3::zero(),
                       L_texture    = (buffer<vec3>*)NULL);

    buffer<float> zbuffer(opt.screen_width, opt.screen_height);

    worker_pool wpool(opt.nthreads);
    if (!wpool.init()) {
        fprintf(stderr, "fatal: failed to initialize threads!\n");
        exit(1);
    }

    context<Uni, VAttr, Px> raster(wpool, initial_uni);

    auto vs = [](const Uni& uni, const vec3& v, const VAttr& attr) {
        // this assumes that the linear part of L_modelview is orthonormal
        vec4 n = uni[L_modelview] * attr[L_normal].v4();
        return (L_pos     = uni[L_projection] * uni[L_modelview] * v.p4(),
                L_normal  = vec3(n.x, n.y, n.z),
                L_qlinear = (L_texcoord = attr[L_texcoord]));
    };

    auto default_shader = PX_MAKE_SHADER_OBJ(raster, vs,
    ([](const Uni& uni, const varying_t& vary, int_4 x, int_4 y,
        float_4& z, simd<Px>::type& color) {
        vec3_4 c;
        float_4 light_val;
        color_and_light(uni, vary, c, light_val);
        color = convert_colors(light_val * c);
    }));

    auto toon_shader = PX_MAKE_SHADER_OBJ(raster, vs,
    ([](const Uni& uni, const varying_t& vary, int_4 x, int_4 y,
        float_4& z, simd<Px>::type& color) {
        vec3_4 c;
        float_4 light_val;
        color_and_light(uni, vary, c, light_val);

        // threshold the light value
        mask_4 hi  = light_val >= float_4(.8);
        mask_4 mid = light_val >= float_4(.4);
        light_val = simd<float>::choose(hi, float_4(1.f),
            simd<float>::choose(mid, float_4(.6f), float_4(.2f)));

        color = convert_colors(light_val * c);
    }));

    auto white_shader = PX_MAKE_SHADER_OBJ(raster, vs,
    ([](const Uni& uni, const varying_t& vary, int_4 x, int_4 y,
        float_4& z, simd<Px>::type& color) {
        color = convert_colors(vec3_4(vec3(1.f, 1.f, 1.f)));
    }));

    shader_obj_id shaders[3] = { default_shader, toon_shader, white_shader };
    int which_shader = 0;
    raster.bind_shader(shaders[which_shader]);

    // want to keep a framerate
    framerate_meter fpsmeter(250);

    // current camera translation
    float cam_z = 2.f;

    for (;;) {
    uint32_t now = fpsmeter.tick(SDL_GetTicks());
    float theta = 2 * float(M_PI) * now / 10000.f;

    SDL_Event ev;
    bool quit = false;
    while (SDL_PollEvent(&ev)) {
        if (ev.type != SDL_KEYDOWN)
            continue;

        if (ev.key.keysym.sym == SDLK_ESCAPE)
            quit = true;
        else if (ev.key.keysym.sym == SDLK_w)
            raster.set_wireframe(!raster.wireframe());
        else if (ev.key.keysym.sym == SDLK_c)
            raster.enable_cull(!raster.cull_enabled());
        else if (ev.key.keysym.sym == SDLK_z)
            raster.enable_depth_write(!raster.depth_write_enabled());
        else if (ev.key.keysym.sym == SDLK_k)
            raster.enable_color_write(!raster.color_write_enabled());
        else if (ev.key.keysym.sym == SDLK_UP)
            cam_z = std::max(.5f, cam_z - .1f);
        else if (ev.key.keysym.sym == SDLK_DOWN)
            cam_z = std::min(100.f, cam_z + .1f);
        else if (ev.key.keysym.sym == SDLK_t) {
            which_shader += 1;
            which_shader %= 3;
            raster.bind_shader(shaders[which_shader]);
        }
    }
    if (quit)
        break;

    // Get access to screen pixels
    if (SDL_MUSTLOCK(screen.get())) {
        if (SDL_LockSurface(screen.get()) < 0) {
            fprintf(stderr, "fatal: unable to lock screen!\n");
            exit(1);
        }
    }

    // projection transformation
    raster.uniform()[L_projection] = mat4::perspective(
        .25f * M_PI, opt.screen_width, opt.screen_height, .01f, 1000.f);

    raster.set_near_distance(.01f);

    // modelview transformation - view part
    mat4 mview = mat4::rotate_x(.25f * M_PI);
    mview = mview * mat4::translate(vec3(0.f, -cam_z, -cam_z));

    // modelview transformation - model part
    //mview = mview * mat4::rotate_z(theta);
    mview = mview * mat4::rotate_y(theta);
    //mview = mview * mat4::rotate_x(theta);

    // "fix" model space -
    // translate mod_center to origin,
    // and scale mod_maxdim to 1.
    mview = mview * mat4::scale(1.f / mod_maxdim);
    mview = mview * mat4::translate(-mod_center);

    raster.uniform()[L_modelview] = mview;

    // set light direction
    vec4 light_dir = mat4::rotate_y(-.75f * M_PI) * vec4::x_basis();
    raster.uniform()[L_light_dir] = light_dir.xyz();

    // draw
    raster.bind_buffers(&target, &zbuffer);
    raster.set_cull_face(CULL_BACK);
    target.clear(0x0000007f);
    zbuffer.clear(std::numeric_limits<float>::infinity());

    raster.uniform()[L_texture] = texture.get();
    raster.draw_tris(mod.nverts, mod.verts, mod.ntris, mod.tris,
                     (L_texcoord = mod.texcoords,
                      L_normal   = mod.normals));

    // show depth buffer if color write is disabled
    if (!raster.color_write_enabled()) {
        float_4 inf(std::numeric_limits<float>::infinity());
        float_4 two(2.f), mins(inf), maxes(0.f), zero(0.f);
        for (int y = 0; y < zbuffer.height(); y++) {
            for (int x = 0; x < zbuffer.width(); x += 4) {
                float_4 depth = zbuffer.at_quad(x, y);
                mask_4  mask  = depth < inf; // was depth written?
                mins.min_d(simd<float>::choose(mask, depth, two));
                maxes.max_d(simd<float>::choose(mask, depth, zero));
            }
        }

        float min = mins.get(0), max = maxes.get(0);
        for (int i = 1; i < 4; i++) {
            min = std::min(min, mins.get(i));
            max = std::max(max, maxes.get(i));
        }

        mins  = float_4(min);
        float_4 scale = float_4(max - min).recip();
        for (int y = 0; y < zbuffer.height(); y++) {
            for (int x = 0; x < zbuffer.width(); x += 4) {
                float_4 depth = zbuffer.at_quad(x, y);
                depth -= mins;
                depth *= scale;
                depth  = float_4(1.f) - depth;
                vec3_4 depth_color(depth, depth, depth);
                target.at_quad(x, y) = convert_colors(depth_color);
            }
        }
    }

    raster.draw_splits(0xff0000);

    // Done rendering, unlock the screen.
    if (SDL_MUSTLOCK(screen.get()))
        SDL_UnlockSurface(screen.get());

    // write FPS to the screen
    int fr = fpsmeter.framerate() + .5f;
    char fr_msg[32];
    snprintf(fr_msg, sizeof (fr_msg), "fps:%d", fr);

#ifndef NO_SDL_TTF
    SDL_Color black = { 0, 0, 0 };
    SDL_Color white_palette[] = { { 0, 0, 0 }, { 0xff, 0xff, 0xff } };
    unique_surface_ptr msg_surface(TTF_RenderText_Solid(
        font.get(), fr_msg, black));
    if (msg_surface == NULL) {
        // non-fatal
        fprintf(stderr, "error: TTF_RenderText_Solid: %s\n", TTF_GetError());
    } else {
        SDL_Rect dst_rect = { .x = 12, .y = 12 }; // w, h ignored
        SDL_BlitSurface(msg_surface.get(), NULL, screen.get(), &dst_rect);
        SDL_SetPalette(msg_surface.get(), SDL_LOGPAL, white_palette, 0, 2);
        dst_rect.x = 10;
        dst_rect.y = 10;
        SDL_BlitSurface(msg_surface.get(), NULL, screen.get(), &dst_rect);
    }
#else
    if (now - last_fps_print >= 250) {
        last_fps_print = now;
        printf("%s\n", fr_msg);
    }
#endif

    // Update the entire screen
    SDL_UpdateRect(screen.get(), 0, 0, opt.screen_width, opt.screen_height);

    } // SDL_GetTicks() loop

    // Done
    return 0;
}

inline uint_4 convert_colors(vec3_4 c) {
    float_4 zero(0.f), one(1.f);

    c.x.clamp_d(zero, one);
    c.y.clamp_d(zero, one);
    c.z.clamp_d(zero, one);

    c *= float_4(255.f);
    uint_4 rgba(0xff);
    rgba <<= 8;
    rgba  |= uint_4(c.x);
    rgba <<= 8;
    rgba  |= uint_4(c.y);
    rgba <<= 8;
    rgba  |= uint_4(c.z);

    return rgba;
}

options parse_options(int argc, char *argv[]) {
    // defaults
    options opt;
    opt.nthreads      = DEFAULT_NTHREADS;
    opt.screen_flags  = DEFAULT_SCREEN_FLAGS;
    opt.screen_width  = DEFAULT_SCREEN_WIDTH;
    opt.screen_height = DEFAULT_SCREEN_HEIGHT;
    opt.font_file     = DEFAULT_FONT_FILE;
    opt.obj_file      = DEFAULT_OBJ_FILE;
    opt.tex_file      = DEFAULT_TEX_FILE;

    // override with switches
    for (int i = 1; i < argc; i++) {
        if (strcmp(argv[i], "-n") == 0) {
            if (++i < argc) {
                char* arg = argv[i];
                opt.nthreads = strtol(arg, &arg, 0);
                if (opt.nthreads < 1 || opt.nthreads > 32 || *arg != '\0') {
                    fprintf(stderr, "error: invalid number of threads %s "
                            "(should be an integer between 1 and 32)\n",
                            argv[i]);
                    exit(2);
                }
            } else {
                fprintf(stderr, "error: -n requires an argument\n");
                exit(2);
            }
        } else if (strcmp(argv[i], "-fs") == 0) {
            opt.screen_flags |= SDL_FULLSCREEN;
        } else if (strcmp(argv[i], "-dim") == 0) {
            if (++i < argc) {
                char* arg = argv[i];
                opt.screen_width = strtol(arg, &arg, 0);
                bool ok = true;
                if (*arg != 'x') {
                    ok = false;
                } else {
                    arg++;
                    opt.screen_height = strtol(arg, &arg, 0);
                    ok = *arg == '\0';
                }

                if (!ok) {
                    fprintf(stderr, "error: invalid dimension %s "
                            "(should be wxh, e.g. 800x600)\n", argv[i]);
                    exit(2);
                } else if (opt.screen_width % 4 != 0) {
                    fprintf(stderr, "error: width %d not a multiple of 4\n",
                            opt.screen_width);
                    exit(2);
                } else if (opt.screen_width < 0 || opt.screen_height < 0) {
                    fprintf(stderr, "error: negative dimension %s\n", argv[i]);
                    exit(2);
                }
            } else {
                fprintf(stderr, "error: -dim requires an argument\n");
                exit(2);
            }
        } else if (strcmp(argv[i], "-fn") == 0) {
            if (++i < argc) {
                opt.font_file = argv[i];
            } else {
                fprintf(stderr, "error: -fn requires an argument\n");
                exit(2);
            }
        } else if (strcmp(argv[i], "-mod") == 0) {
            if (++i < argc) {
                opt.obj_file = argv[i];
            } else {
                fprintf(stderr, "error: -mod requires an argument\n");
                exit(2);
            }
        } else if (strcmp(argv[i], "-tex") == 0) {
            if (++i < argc) {
                opt.tex_file = argv[i];
            } else {
                fprintf(stderr, "error: -tex requires an argument\n");
                exit(2);
            }
        } else {
            fprintf(stderr, "error: unrecognized argument %s\n", argv[i]);
            exit(2);
        }
    }

    return opt;
}
